# -*- coding: utf-8 -*-

from GAparsimony.lhs.util import findorder
from GAparsimony import order
import numpy as np

   
def randomLHS(n, k, bPreserveDraw=False, seed=None):
    r"""Construct a random Latin hypercube design

    `randomLHS(4,3)` returns a 4x3 matrix with each column constructed as follows: A random per-mutation of (1,2,3,4) 
    is generated, say (3,1,2,4) for each of K columns. Then a uniform randomnumber is picked from each indicated 
    quartile. In this example a random number between `0.5` and `0.75` is chosen, then one between `0 `and `0.25`, then one 
    between `0.25` and `0.5`, finally one between `0.75` and `1`.

    Parameters
    ----------
    n : int
        The number of rows or samples. 
    k : int
        The number of columns or parameters/variables.
    bPreserveDraw : bool, optional
        Should the draw be constructed so that it is the same for variable numbers of columns?. Default `False`
    seed : int, optional
        Random seed. Default `None`.

    Returns
    -------
    numpy.array
        A `numpy.array` of `float` with shape `(n, k)`.
    """
    
    if n < 1 or k < 1:
        raise Exception("nsamples are less than 1 (n) or nparameters less than 1 (k)")

    if seed:
        np.random.seed(seed)

    result = np.zeros(n*k).reshape((n, k))
        
    randomunif1 = np.empty(n).astype(np.double)

    if bPreserveDraw:
        
        randomunif2 = np.empty(n).astype(np.double)
        for jcol in range(k):
            
            for irow in range(n):
                randomunif1[irow] = np.random.uniform(low=0, high=1)
            for irow in range(n):
                randomunif2[irow] = np.random.uniform(low=0, high=1)


            orderVector = order(randomunif1)
            for irow in range(n):
                result[irow,jcol] = orderVector[irow] + randomunif2[irow]
                result[irow,jcol] =  result[irow,jcol] / np.double(n)

    else:
        randomunif2 = np.empty(n*k).astype(np.double)
        for jcol in range(k):
            for irow in range(n):
                randomunif1[irow] = np.random.uniform(low=0, high=1)

            orderVector = order(randomunif1)
            for irow in range(n):
                result[irow,jcol] = orderVector[irow]

        for i in range(n*k):
            randomunif2[i] = np.random.uniform(low=0, high=1)

        randomunif2 = randomunif2.reshape((n, k))
        for jcol in range(k):

            for irow in range(n):
                result[irow,jcol] = result[irow,jcol] + randomunif2[irow, jcol]
                result[irow,jcol] = result[irow,jcol] / np.double(n)

    return result

def randomLHS_int(n, k, seed=None):
    r"""Construct a random Latin hypercube design

    `randomLHS(4,3)` returns a 4x3 matrix with each column constructed as follows: A random per-mutation of (1,2,3,4) 
    is generated, say (3,1,2,4) for each of K columns. Then a uniform randomnumber is picked from each indicated 
    quartile. In this example a random number between `0.5` and `0.75` is chosen, then one between `0 `and `0.25`, then one 
    between `0.25` and `0.5`, finally one between `0.75` and `1`.

    Parameters
    ----------
    n : int
        The number of rows or samples. 
    k : int
        The number of columns or parameters/variables.
    seed : int, optional
        Random seed. Default `None`.

    Returns
    -------
    numpy.array
        A `numpy.array` of `int` with shape `(n, k)`.
    """

    if seed:
        np.random.seed(seed)

    result = np.empty((n, k)).astype(np.int32)
    randomunif1 = np.empty(n).astype(np.double)
    for jcol in range(k):
    
        for irow in range(n):
            randomunif1[irow] = np.random.uniform(low=0, high=1)
        
        orderVector = findorder(randomunif1)
        for irow in range(n):
            result[irow,jcol] = orderVector[irow]

    return result



# -*- coding: utf-8 -*-

import numpy as np
from GAparsimony.lhs.util import initializeAvailableMatrix, convertIntegerToNumericLhs

# result es lo que devolvemos

def _maximinLHS(n, k, dup=1, seed=None):

    if n < 1 or k < 1 or dup < 1:
        raise Exception("nsamples are less than 1 (n) or nparameters less than 1 (k) or duplication is less than 1")

    if seed:
        np.random.seed(seed)

    nsamples = n
    nparameters = k
    duplication = dup

    result = np.empty((k, n)).astype(np.double)

    # the length of the point1 columns and the list1 vector 
    leng = duplication * (nsamples - 1)

    # create memory space for computations 
    point1 = np.empty((nparameters, leng))
    list1 = np.empty(leng)
    vec = np.empty(nparameters)

    #  squared distance between corner (1,1,1,..) and (N,N,N,...) 
    squaredDistanceBtwnCorners = np.double(nparameters * (nsamples - 1) * (nsamples - 1))
    avail = initializeAvailableMatrix(nparameters, nsamples)

    # * come up with an array of K integers from 1 to N randomly
    # * and put them in the last column of result
    for irow in range(nparameters):
        result[irow, nsamples-1] = np.double(np.floor(np.random.uniform(low=0, high=1) * np.double(nsamples) + 1.0))
    
    # * use the random integers from the last column of result to place an N value
    # * randomly through the avail matrix
    # 
    for irow in range(nparameters):
        avail[irow,int(result[irow, nsamples - 1] - 1)] = int(nsamples)
    
    #  move backwards through the result matrix columns 
    for ucount in range(nsamples - 1, 0, -1):
        for irow in range(nparameters):
            for jcol in range(duplication):
                #  create the list1 vector 
                for j in range(ucount):              
                    list1[j + ucount*jcol] = avail[irow, j]
                
            #  create a set of points to choose from 
            for jcol in range(ucount * duplication, 0, -1):
                point_index = int(np.floor(np.random.uniform(low=0, high=1) * np.double(jcol)))
                point1[irow, jcol-1] = list1[point_index]
                list1[point_index] = list1[jcol - 1]
            
        minSquaredDistBtwnPts = np.finfo(np.double).tiny #DBL_MIN el menor float representable
        best = 0
        for jcol in range(duplication * ucount - 1):
            #  set min candidate equal to the maximum distance to start 
            minCandidateSquaredDistBtwnPts = int(np.ceil(squaredDistanceBtwnCorners))
            for  j in range(ucount, nsamples):
                distSquared = 0

                # * find the distance between candidate points and the points already
                # * in the sample
                for kindex in range(nparameters):
                    vec[kindex] = point1[kindex, jcol] - result[kindex, j]
                    distSquared = distSquared + (vec[kindex] * vec[kindex])
                
                # * if the distance squared value is the smallest so far, place it in the
                # * min candidate
                if minCandidateSquaredDistBtwnPts > distSquared:
                    minCandidateSquaredDistBtwnPts = distSquared
                
            # * if the candidate point is the largest minimum distance between points so
            # * far, then keep that point as the best.
            if np.double(minCandidateSquaredDistBtwnPts) > minSquaredDistBtwnPts:
                minSquaredDistBtwnPts = np.double(minCandidateSquaredDistBtwnPts)
                best = int(jcol)
            
        #  take the best point out of point1 and place it in the result 
        for irow in range(nparameters):
            result[irow, ucount-1] = point1[irow, best]

        #  update the numbers that are available for the future points 
        for irow in range(nparameters):
            for jcol in range(nsamples):
                if avail[irow, jcol] == result[irow, ucount-1]:
                    avail[irow, jcol] = avail[irow, ucount-1]
                
    # * once all but the last points of result are filled in, there is only
    # * one choice left
    for irow in range(nparameters):
        result[irow, 0] = avail[irow, 0]

    return result.transpose()


def maximinLHS(n, k, dup=1, seed=None):
    r"""Maximin Latin Hypercube Sample

    Draws a Latin Hypercube Sample from a set of uniform distributions for use in creating 
    a Latin Hypercube Design. This function attempts to optimize the sample by 
    maximizing the minium distance between design points (maximin criteria).

    Parameters
    ----------
    n : int
        The number of rows or samples. 
    k : int
        The number of columns or parameters/variables.
    dup : int, optional
        A factor that determines the number of candidate points used in the search. 
        A multiple of the number of remaining points than can be added. Default `1`.
    seed : int, optional
        Random seed. Default `None`.

    Returns
    -------
    numpy.array
        An `n` by `n` Latin Hypercube Sample matrix with values uniformly distributed on `[0,1]`.
    """
    
    return np.random.rand(k)[np.newaxis, :] if n==0 else convertIntegerToNumericLhs(_maximinLHS(n, k, dup, seed))



# -*- coding: utf-8 -*-

import numpy as np

class Population:

    INTEGER = 0
    FLOAT = 1
    CATEGORICAL = 2
    CONSTANT = 3

    def __init__(self, params, columns, population = None):
        r"""
        This class is used to create the GA populations. 
        Allow chromosomes to have int, float, and constant values. 


        Parameters
        ----------
        params : dict
            It is a dictionary with the model's hyperparameters to be adjusted and the search space of them.
            
            .. code-block::

                {
                    "<< hyperparameter name >>": {
                        "range": [<< minimum value >>, << maximum value >>],
                        "type": GAparsimony.FLOAT/GAparsimony.INTEGER
                    },
                    "<< hyperparameter name >>": {
                        "value": << constant value >>,
                        "type": GAparsimony.CONSTANT
                    }
                }

        columns : int or list of str
            The number of features/columns in the dataset or a list with their names.
        population : numpy.array, optional
            It is a float matrix that represents the population. Default `None`.

        Attributes
        ----------
        population : Population
            The population.
        _min : numpy.array
            A vector of length `params+columns` with the smallest values that can take.
        _max : numpy.array
            A vector of length `params+columns` with the highest values that can take.
        _params : dict
            Dict with the parameter values.
        const : dict
            Dict with the constants values.
        colsnames : list of str
            List with the columns names.
        """
        
        if type(params) is not dict:
            raise Exception("params must be of type dict !!!")

        self._min = np.array([(0 if params[x]["type"] is Population.CATEGORICAL else params[x]["range"][0]) for x in params if params[x]["type"] is not Population.CONSTANT])
        self._max = np.array([(len(params[x]["range"]) if params[x]["type"] is Population.CATEGORICAL else params[x]["range"][1]) for x in params if params[x]["type"] is not Population.CONSTANT])

        self._params = dict((x, params[x]) for x in params if params[x]["type"] is not Population.CONSTANT)
        self.const = dict((x, params[x]["value"]) for x in params if params[x]["type"] is Population.CONSTANT)

        self.colsnames = (columns if type(columns) is list else columns.tolist()) if hasattr(columns, '__iter__') else [f"col_{i}" for i in range(columns)]
        self._min = np.concatenate((self._min, np.zeros(len(self.colsnames))), axis=0)
        self._max = np.concatenate((self._max, np.ones(len(self.colsnames))), axis=0)

        self._pos_n, self._pos_c = list(), list()
        for i, x in enumerate(self._params):
            if self._params[x]["type"] is Population.CATEGORICAL:
                self._pos_c.append(i)
            else:
                self._pos_n.append(i)

        def _trans_mut():
            t = list()
            gen = list()
            for x in self.paramsnames:
                if params[x]["type"] == Population.INTEGER:
                    t.append(np.vectorize(lambda x: int(x), otypes=[int]))
                    gen.append(lambda y, x=x, **kwargs: np.random.randint(low=self._min[y], high=self._max[y]))
                elif params[x]["type"] == Population.FLOAT:
                    t.append(np.vectorize(lambda x: float(x), otypes=[float]))
                    gen.append(lambda y, x=x, **kwargs: np.random.uniform(low=self._min[y], high=self._max[y]))
                elif params[x]["type"] == Population.CATEGORICAL:
                    t.append(np.vectorize(lambda y, x=x: y if type(y) is str else params[x]["range"][int(np.trunc(y))], otypes=[str]))
                    gen.append(lambda y, x=x, **kwargs: np.random.randint(low=self._min[y], high=self._max[y]))
            t.extend([lambda x: x>0.5]*len(self.colsnames))
            gen.extend([lambda y, x=x, **kwargs: np.random.uniform(low=self._min[y], high=self._max[y]) <= kwargs["feat_mut_thres"]]*len(self.colsnames))

            def aux(x):
                if len(x.shape)>1:
                    return np.array(list(map(lambda f, c: f(x[:, c]), t, range(0, x.shape[1]))), dtype=object).T
                else:
                    return list(map(lambda f, c: f(c), t, x))

            return aux, gen


        self._transformers, self.random_gen = _trans_mut()
        

        if population is not None:
            if type(population) is not np.ndarray or len(population.shape) < 2:
                raise Exception("Popularion is not a numpy matrix")
            self.population = population

    @property
    def population(self):
        return self._transformers(self._pop)

    @population.setter
    def population(self, population):
        self._pop = np.apply_along_axis(lambda x: x.astype(object), 1, population.astype(object))

    @property
    def paramsnames(self):
        return list(self._params.keys())

    def __getitem__(self, key):
        return self._pop[key]

    def __setitem__(self, key, newvalue):
        self._pop[key] = newvalue

    def getChromosome(self, key):
        r"""
        This method returns a chromosome from the population. 

        Parameters
        ----------
        key : int
            Chromosome row index .

        Returns
        -------
        Chromosome
            A `Chromosome` object.
        """
        data = self._transformers(self._pop[key, :])
        return Chromosome(data[:len(self.paramsnames)], self.paramsnames, self.const, data[len(self.paramsnames):], self.colsnames)


    
class Chromosome:

    # @autoassign
    def __init__(self, params, name_params, const, cols, name_cols):
        r"""
        This class defines a chromosome which includes the hyperparameters, the constant values, and the feature selection.


        Parameters
        ----------
        params : numpy.array
            The algorithm hyperparameter values.
        name_params : list of str
            The names of the hyperparameters.
        const : numpy.array
            A dictionary with the constants to include in the chomosome.
        cols : numpy.array
            The probabilities for selecting the input features (selected if prob>0.5).
        name_cols : list of str
            The names of the input features.

        Attributes
        ----------
        params : dict
            A dictionary with the parameter values (hyperparameters and constants).
        columns : numpy.array of bool
            A boolean vector with the selected features.
        """
        self._params = params
        self.name_params = name_params
        self.const = const
        self._cols = cols
        self.name_cols = name_cols

    @property
    def params(self):
        return {**dict(zip(self.name_params, self._params)), **self.const}

    @property
    def columns(self):
        return self._cols



# -*- coding: utf-8 -*-

import numpy as np
from GAparsimony.lhs.util import findorder, sumInvDistance, convertIntegerToNumericLhs


def _optimumLHS(n, k, optimalityRecordLength, maxSweeps=2, eps=0.1, seed=None):

    if seed:
        np.random.seed(seed)

    nOptimalityRecordLength = optimalityRecordLength
    nsamples = n
    nparameters = k
    nMaxSweeps = maxSweeps
    eps_change = eps
    optimalityChangeOld = 0.0
    outlhs = np.empty((nsamples, nparameters))
    optimalityRecord = np.empty(nOptimalityRecordLength).astype(np.double)
    interchangeRow1 = np.zeros(nOptimalityRecordLength)
    interchangeRow2 = np.zeros(nOptimalityRecordLength)

    #  fill the oldHypercube with a random lhs sample
    randomUnif = np.empty(nsamples).astype(np.double)
    for jcol in range(nparameters):
        #  fill a vector with a random sample to order
        for irow in range(nsamples):
            randomUnif[irow] = np.random.uniform(low=0, high=1)
        
        orderedUnif = findorder(randomUnif)
        for irow in range(nsamples):
            outlhs[irow,jcol] = orderedUnif[irow]
    
    # find the initial optimality measure 
    gOptimalityOld = sumInvDistance(outlhs)
    test = 0
    iterator = 0
    while test == 0:
    
        if iterator == nMaxSweeps:
            break
        iterator += 1
        # iterate over the columns 
        for j in range(nparameters):
        
            optimalityRecordIndex = 0
            # iterate over the rows for the first point from 0 to N-2 
            for i in range(nsamples - 1):
            
                # iterate over the rows for the second point from i+1 to N-1 
                for kindex in range((i + 1), nsamples):
                
                    # put the values from oldHypercube into newHypercube 
                    newHypercube = outlhs.copy()
                    # exchange two values (from the ith and kth rows) in the jth column
                    # * and place them in the new matrix 
                    newHypercube[i, j] = outlhs[kindex, j]
                    newHypercube[kindex, j] = outlhs[i, j]
                    # store the optimality of the newly created matrix and the rows that
                    # * were interchanged 
                    optimalityRecord[optimalityRecordIndex] = sumInvDistance(newHypercube)
                    interchangeRow1[optimalityRecordIndex] = i
                    interchangeRow2[optimalityRecordIndex] = kindex
                    optimalityRecordIndex += 1

            # once all combinations of the row interchanges have been completed for
            # * the current column j, store the old optimality measure (the one we are
            # * trying to beat) 
            optimalityRecord[optimalityRecordIndex] = gOptimalityOld
            interchangeRow1[optimalityRecordIndex] = 0
            interchangeRow2[optimalityRecordIndex] = 0
            # Find which optimality measure is the lowest for the current column.
            # * In other words, which two row interchanges made the hypercube better in
            # * this column 
            posit = 0
            for kindex in range(nOptimalityRecordLength):
            
                if optimalityRecord[kindex] < optimalityRecord[posit]:
                    posit = kindex
            # If the new minimum optimality measure is better than the old measure 
            if optimalityRecord[posit] < gOptimalityOld:
            
                # put oldHypercube in newHypercube 
                newHypercube = outlhs.copy()
                # Interchange the rows that were the best for this column 
                newHypercube[int(interchangeRow1[posit]), j] = outlhs[int(interchangeRow2[posit]), j]
                newHypercube[int(interchangeRow2[posit]), j] = outlhs[int(interchangeRow1[posit]), j]
                # put newHypercube back in oldHypercube for the next iteration 
                outlhs = newHypercube.copy()
                # if this is not the first column we have used for this sweep 
                if j > 0:
                
                    # check to see how much benefit we gained from this sweep 
                    optimalityChange = np.abs(optimalityRecord[posit] - gOptimalityOld)
                    if optimalityChange < eps_change * optimalityChangeOld:
                        test = 1
                # if this is first column of the sweep, then store the benefit gained 
                else:
                
                    optimalityChangeOld = np.abs(optimalityRecord[posit] - gOptimalityOld)
                
                # replace the old optimality measure with the current one 
                gOptimalityOld = optimalityRecord[posit]
            
            # if the new and old optimality measures are equal 
            elif optimalityRecord[posit] == gOptimalityOld:
                test = 1
            
            # if the new optimality measure is worse 
            elif optimalityRecord[posit] > gOptimalityOld:
                raise Exception("Unexpected Result: Algorithm produced a less optimal design")
            if test == 1:
                break
    return outlhs




def optimumLHS(n, k, maxsweeps=2, eps=0.1, seed=None):
    r"""Optimum Latin Hypercube Sample

    Draws a Latin Hypercube Sample from a set of uniform distributions for use in creating a LatinHypercube Design. 
    This function uses the Columnwise Pairwise (CP) algorithm to generate an optimal design with respect to the S optimality criterion.

    Parameters
    ----------
    n : int
        The number of partitions (simulations or design points or rows). 
    k : int
        The number of replications (variables or columns).
    maxsweeps : int, optional
        The maximum number of times the CP algorithm is applied to all the columns. Default `2`
    eps : float, optional
        The optimal stopping criterion. Algorithm stops when the change in optimality
        measure is less than `eps*100%` of the previous value. Default `0.01`
    seed : int, optional
        Random seed. Default `None`.

    Returns
    -------
    numpy.array
        An `n` by `n` Latin Hypercube Sample matrix with values uniformly distributed on `[0,1]`.
    """

    if seed:
        np.random.seed(seed)
    jLen = int(np.math.factorial(n)/(np.math.factorial(2)*np.math.factorial(n-2)) + 1.0)
    return np.random.rand(k)[np.newaxis, :] if n==0 else convertIntegerToNumericLhs(_optimumLHS(n, k, jLen, maxsweeps, eps, seed))




# -*- coding: utf-8 -*-

"""Complexity module.

This module contains predefined complexity functions for some of the most popular algorithms in the scikit-learn library:

* **linearModels_complexity**: Any algorithm from `sklearn.linear_model'. Returns: 10^9·nFeatures + (sum of the squared coefs).
* **svm_complexity**: Any algorithm from `sklearn.svm'. Returns: 10^9·nFeatures + (number of support vectors).
* **knn_complexity**: Any algorithm from `sklearn.neighbors'. Returns: 10^9·nFeatures + 1/(number of neighbors)
* **mlp_complexity**: Any algorithm from `sklearn.neural_network'. Returns: 10^9·nFeatures + (sum of the ANN squared weights).
* **randomForest_complexity**: Any algorithm from `sklearn.ensemble.RandomForestRegressor' or 'sklearn.ensemble.RandomForestClassifier'. Returns: 10^9·nFeatures + (the average of tree leaves).
* **xgboost_complexity**: XGboost sklearn model. Returns: 10^9·nFeatures + (the average of tree leaves * number of trees) (Experimental)
* **decision_tree_complexity**: Any algorithm from 'sklearn.tree'. Return: 10^9·nFeatures + (number of leaves) (Experimental)

Otherwise:

* **generic_complexity**: Any algorithm. Returns: the number of input features (nFeatures).

Other complexity functions can be defined with the following interface.

.. highlight:: python
.. code-block:: python

    def complexity(model, nFeatures, **kwargs):
        pass
            
    return complexity
"""

import numpy as np

def generic_complexity(model, nFeatures, **kwargs):
    r"""
    Generic complexity function.

    Parameters
    ----------
    model : model
        The model for calculating complexity.
    nFeatures : int
        The number of input features the model has been trained with.
    **kwargs : 
        A variable number of named arguments.

    Returns
    -------
    int
        nFeatures.

    """
    return nFeatures

def linearModels_complexity(model, nFeatures, **kwargs):
    r"""
    Complexity function for linear models.

    Parameters
    ----------
    model : model
        The model for calculating complexity.
    nFeatures : int
         The number of input features the model has been trained with.
    **kwargs : 
        A variable number of named arguments.

    Returns
    -------
    int
        10^9·nFeatures + (sum of the model squared coefs).

    """

    int_comp = np.min((1E09-1,np.sum(model.coef_**2))) # Internal Complexity Sum of squared weigths
    return nFeatures*1E09 + int_comp

def svm_complexity(model, nFeatures, **kwargs):
    r"""
    Complexity function for SVM models.

    Parameters
    ----------
    model : model
        The model for calculating complexity.
    nFeatures : int
         The number of input features the model has been trained with.
    **kwargs : 
        A variable number of named arguments.

    Returns
    -------
    int
        10^9·nFeatures + (number of support vectors)

    """

    int_comp = np.min((1E09-1,np.sum(model.n_support_))) # Internal Complexity
    return nFeatures*1E09 + int_comp

def knn_complexity(model, nFeatures, **kwargs):
    r"""
    Complexity function for KNN models.

    Parameters
    ----------
    model : model
        The model for calculating complexity.
    nFeatures : int
         The number of input features the model has been trained with.
    **kwargs : 
        A variable number of named arguments.

    Returns
    -------
    int
        10^9·nFeatures + 1/(number of neighbors)

    """

    int_comp = 1E06 * 1/model.n_neighbors   # More k less flexible
    return nFeatures*1E09 + int_comp

def mlp_complexity(model, nFeatures, **kwargs):
    r"""
    Complexity function for MLP models.

    Parameters
    ----------
    model : model
        The model for calculating complexity.
    nFeatures : int
         The number of input features the model has been trained with.
    **kwargs : 
        A variable number of named arguments.

    Returns
    -------
    int
        10^9·nFeatures + (sum of the ANN squared weights)

    """

    weights = [np.concatenate(model.intercepts_)]
    for wm in model.coefs_:
        weights.append(wm.flatten())
    weights = np.concatenate(weights) 
    int_comp = np.min((1E09-1,np.sum(weights**2)))
    return nFeatures*1E09 + int_comp

def randomForest_complexity(model, nFeatures, **kwargs):
    r"""
    Complexity function for Random Forest models.

    Parameters
    ----------
    model : model
        The model for calculating complexity.
    nFeatures : int
        The number of input features the model has been trained with.
    **kwargs : 
        A variable number of named arguments.

    Returns
    -------
    int
        10^9·nFeatures + (the average of tree leaves)

    """

    num_leaves = [tree.get_n_leaves() for tree in model.estimators_]
    int_comp = np.min((1E09-1,np.mean(num_leaves))) # More leaves more complex  
    return nFeatures*1E09 + int_comp

def xgboost_complexity(model, nFeatures, **kwargs):
    r"""
    Complexity function for XGBoost model.

    Parameters
    ----------
    model : model
        The model for calculating complexity.
    nFeatures : int
         The number of input features the model has been trained with.
    **kwargs : 
        A variable number of named arguments.

    Returns
    -------
    int
        10^9·nFeatures + (the average of tree leaves * number of trees) (Experimental)

    """
    df_model = model.get_booster().trees_to_dataframe()
    df_model = df_model[df_model.Feature=='Leaf']
    mean_leaves = df_model.groupby('Tree')['Feature'].count().mean()
    num_trees = df_model.Tree.nunique()   
    int_comp = np.min((1E09-1,num_trees*mean_leaves))
    return nFeatures*1E09 + int_comp


def decision_tree_complexity(model, nFeatures, **kwargs):
    r"""
    Complexity function for decision tree model.

    Parameters
    ----------
    model : model
        The model for calculating complexity.
    nFeatures : int
         The number of input features the model has been trained with.
    **kwargs : 
        A variable number of named arguments.

    Returns
    -------
    int
        10^9·nFeatures + (number of leaves)

    """
    num_leaves = model.get_n_leaves()
    int_comp = np.min((1E09-1,num_leaves)) # More leaves more complex  
    return nFeatures*1E09 + int_comp




# -*- coding: utf-8 -*-

import numpy as np
from GAparsimony.lhs.util import initializeAvailableMatrix, convertIntegerToNumericLhs


def _improvedLHS( n,  k, dup=1, seed=None):

    if seed:
        np.random.seed(seed)
    nsamples = n
    nparameters = k
    duplication = dup
    result = np.empty((k, n))
    #  ********** matrix_unsafe<int> m_result = matrix_unsafe<int>(nparameters, nsamples, result);
    #  the length of the point1 columns and the list1 vector 
    size = duplication * (nsamples - 1)
    
    #  create memory space for computations 
    point1 = np.empty((nparameters, size))
    list1 = np.empty(size)
    vec = np.empty(nparameters)
    #  optimum spacing between points 
    opt = np.double(nsamples) / ( np.power(np.double(nsamples), (1.0 / np.double(nparameters))))
    #  the square of the optimum spacing between points 
    opt2 = opt * opt
    #  initialize the avail matrix 
    avail = initializeAvailableMatrix(nparameters, nsamples)

    # * come up with an array of K integers from 1 to N randomly
    # * and put them in the last column of result
    for irow in range(nparameters):
        result[irow, nsamples-1] = int(np.floor(np.random.uniform(low=0, high=1) * np.double(nsamples) + 1.0))
    
    # * use the random integers from the last column of result to place an N value
    # * randomly through the avail matrix
    for irow in range(nparameters):
        avail[irow, int(result[irow, nsamples-1] - 1)] = nsamples
    
    #  move backwards through the result matrix columns.
    for ucount in range(nsamples - 1, 0, -1):
        # unsigned int ucount = static_cast<unsigned int>(count);
        for irow in range(nparameters):
            for jcol in range(duplication):
                #  create the list1 vector 
                for j in range(ucount):
                    list1[j + ucount*jcol] = avail[irow, j]
                
            #  create a set of points to choose from. Note, need to use int
            #  Note: can't do col = count*duplication - 1; col >= 0 because it throws a warning at W4 
            for ujcol in range(ucount * duplication, 0, -1):
                # unsigned int ujcol = static_cast<unsigned int>(jcol);
                point_index = int(np.floor(np.random.uniform(low=0, high=1) * np.double(ujcol)))
                point1[irow, ujcol-1] = list1[point_index]
                list1[point_index] = list1[ujcol-1]
        
        min_all = np.finfo(np.double).tiny
        best = 0
        for jcol in range(duplication * ucount - 1):
            min_candidate = 4294967295
            for j in range(nsamples):
                distSquared = 0
        
                # * find the distance between candidate points and the points already
                # * in the sample
                for kindex in range(nparameters):
                    vec[kindex] = point1[kindex, jcol] - result[kindex, j]
                    distSquared += (vec[kindex] + 10e-10)**2 # vec[kindex] * vec[kindex]   # Da un warning
                
                # * if the distSquard value is the smallest so far place it in
                # * min candidate
                if min_candidate > distSquared:
                
                    min_candidate = distSquared
                
            # * if the difference between min candidate and opt2 is the smallest so
            # * far, then keep that point as the best.
            if np.abs(np.double(min_candidate) - opt2) < min_all:
                min_all = np.abs(np.double(min_candidate) - opt2)
                best = jcol
        
        #  take the best point out of point1 and place it in the result 
        for irow in range(nparameters):
            result[irow, ucount - 1] = point1[irow, best]
        
        #  update the numbers that are available for the future points 
        for irow in range(nparameters):
            for jcol in range(nsamples):
                if avail[irow, jcol] == result[irow, ucount - 1]:
                    avail[irow, jcol] = avail[irow, ucount-1]
                
    # * once all but the last points of result are filled in, there is only
    # * one choice left
    for jrow in range(nparameters):
        result[jrow, 0] = avail[jrow, 0]
    
    
    return result.transpose()

def improvedLHS(n, k, dup=1, seed=None):
    r"""Improved Latin Hypercube Sample

    Draws a Latin Hypercube Sample from a set of uniform distributions for use in creating a Latin
    Hypercube Design. This function attempts to optimize the sample with respect to 
    an optimum euclidean distance between design points.

    Parameters
    ----------
    n : int
        The number of rows or samples. 
    k : int
        The number of columns or parameters/variables.
    dup : int, optional
        A factor that determines the number of candidate points used in the search. 
        A multiple of the number of remaining points than can be added. Default `1`.
    seed : int, optional
        Random seed. Default `None`.

    Returns
    -------
    numpy.array
        A `numpy.array` of `float` with shape `(n, k)`.
    """

    return np.random.rand(k)[np.newaxis, :] if n == 1 else convertIntegerToNumericLhs(_improvedLHS(n, k, dup, seed))


# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd

def order(obj, kind='heapsort', decreasing = False, na_last = True):
    r"""Function to order vectors 

    This function is an overload of `numpy.argsort` sorting method allowing increasing 
    and decreasing ordering and allowing nan values to be placed at the end and at the beginning.

    Parameters
    ----------
    obj : numpy.array
        Array to order.
    kind : {'quicksort', 'mergesort', 'heapsort', 'stable'}, optional
        Sorting algorithm. The default is `heapsort`. Note that both 'stable' and 'mergesort' use timsort 
        under the covers and, in general, the actual implementation will vary with data type.
    decreasing : bool, optional
        If we want decreasing order.
    na_last : bool, optional
        For controlling the treatment of `NA's`. If `True`, missing values in the data are put last, if `False`, they are put first.
    """

    if not decreasing:
        if na_last:
            return obj.argsort(kind=kind)
        else:
            na = np.count_nonzero(np.isnan(obj))
            aux = obj.argsort(kind=kind)
            return np.concatenate([aux[-na:], aux[:-na]])
    else:
        if not na_last:
            return obj.argsort(kind=kind)[::-1]
        else:
            na = np.count_nonzero(pd.isnull(obj))
            aux = obj.argsort(kind=kind)[::-1]
            return np.concatenate([aux[na:], aux[:na]])


__all__ = ['base']

from .base.randomLHS import randomLHS, randomLHS_int
from .base.geneticLHS import geneticLHS
from .base.improvedLHS import improvedLHS
from .base.maximinLHS import maximinLHS
from .base.optimumLHS import optimumLHS



__all__ = ["bclib", "utilityLHS"]

from .bclib import findorder_zero, findorder, inner_product
from.utilityLHS import isValidLHS_int, isValidLHS, initializeAvailableMatrix, runif_std, convertIntegerToNumericLhs, sumInvDistance, calculateDistanceSquared, calculateDistance, calculateSOptimal, runifint


__all__ = ["fitness", "order", "population", "parsimony_monitor", "complexity", "config"]

from .fitness import getFitness
from .population import Population
from .parsimony_monitor import parsimony_monitor, parsimony_summary
from .order import order
from .complexity import generic_complexity, linearModels_complexity, svm_complexity, knn_complexity, mlp_complexity, randomForest_complexity, xgboost_complexity
from .config import CLASSIF_LOGISTIC_REGRESSION, CLASSIF_PERCEPTRON, REG_LASSO, REG_RIDGE, CLASSIF_SVC, REG_SVR, CLASSIF_KNEIGHBORSCLASSIFIER, REG_KNEIGHBORSREGRESSOR , CLASSIF_MLPCLASSIFIER, REG_MLPREGRESSOR, CLASSIF_RANDOMFORESTCLASSIFIER, REG_RANDOMFORESTREGRESSOR


# -*- coding: utf-8 -*-

import numpy as np
from GAparsimony import GAparsimony, Population

import pytest
from .utilTest import autoargs, readJSONFile


# Clase generica, permite instanciarla con diferentes atributos.
class GenericClass(object):
    @autoargs()
    def __init__(self,**kawargs):
        pass


#################################################
#***************TEST POPULATION*****************#
#################################################

@pytest.mark.parametrize("population", [(readJSONFile('./test/outputs/population.json'))])
def test_GAParsimony_regression_boston_population(population):

    pop = Population(population["params"], columns=population["features"])
    
    model = GenericClass(popSize=population["popSize"], seed_ini=population["seed"], feat_thres=population["feat_thres"], population=pop)

    pop.population = GAparsimony._population(model, type_ini_pop="improvedLHS")
    
    assert (pop.population==np.array(population["population_resultado"])).all()


data = readJSONFile('./test/outputs/populationClass.json')
population = Population(data["params"], data["features"], np.array(data["population"]))

@pytest.mark.parametrize("population, slice, value, resultado", 
                        [(population,(slice(2), slice(None)), np.arange(20), np.array(data["population_1"], dtype=object)),
                        (population,(slice(2), slice(None)), np.array([np.arange(20), np.arange(1, 21)]), np.array(data["population_2"], dtype=object)),
                        (population,(slice(2), slice(None)), 0, np.array(data["population_3"], dtype=object)),
                        (population,(1, slice(2)), 1, np.array(data["population_4"], dtype=object)),
                        (population,(1, slice(None)), np.arange(20), np.array(data["population_5"], dtype=object)),
                        (population,(1, slice(2)), np.array([2,2]), np.array(data["population_6"], dtype=object)),
                        (population,(slice(None), 2), 1, np.array(data["population_7"], dtype=object)),
                        (population,(slice(None), 6), 87, np.array(data["population_8"], dtype=object)),
                        (population,(slice(None), 7), 98, np.array(data["population_9"], dtype=object))])
def test_GAParsimony_regression_population_class(population, slice, value, resultado):

    population[slice] = value
    
    assert (population.population==resultado).all()

#################################################
#*****************TEST RERANK*******************#
#################################################

@pytest.mark.parametrize("rerank", [readJSONFile('./test/outputs/rerank.json')])
def test_GAParsimony_regression_boston_rerank(rerank):

    
    model = GenericClass(fitnessval=np.array(rerank["fitnessval"]), complexity=np.array(rerank["complexity"]), 
                        best_score=rerank["best_score"], popSize=rerank["popSize"], rerank_error = 0.01, verbose=0, 
                        population=Population(rerank["params"], rerank["popSize"]))

    result = GAparsimony._rerank(model)
    
    assert (result==np.array(rerank["position"])).all()


#################################################
#****************TEST SELECTION*****************#
#################################################

@pytest.mark.parametrize("selection", [readJSONFile('./test/outputs/selection.json')])
def test_GAParsimony_regression_boston_selection(selection):
    np.random.seed(selection["seed"])
    population=np.array(selection["population"])
    fitnessval=np.array(selection["fitnessval"])
    fitnesstst=np.array(selection["fitnesstst"])
    complexity=np.array(selection["complexity"])
    
    model = GenericClass(selection=selection["selection"], popSize=selection["popSize"], 
                        sel=selection["sel"], population=Population(selection["params"], selection["features"], np.array(selection["population"])), fitnessval=fitnessval,
                        complexity=complexity, fitnesstst=fitnesstst
                        )

    GAparsimony._selection(model)
    
    assert (model.population.population==population[selection["sel"]]).all() and (model.fitnessval==fitnessval[selection["sel"]]).all() and \
            (model.fitnesstst==fitnesstst[selection["sel"]]).all() and (model.complexity==complexity[selection["sel"]]).all()

#################################################
#****************TEST MUTATION*****************#
#################################################

@pytest.mark.parametrize("mutation", [readJSONFile('./test/outputs/mutation.json')])
def test_GAParsimony_regression_boston_mutation(mutation):
    np.random.seed(mutation["seed"])
    
    model = GenericClass(pmutation=mutation["pmutation"], popSize=mutation["popSize"], not_muted=mutation["not_muted"], 
                        population=Population(mutation["params"], mutation["features"], np.array(mutation["population"])), 
                        feat_mut_thres=mutation["feat_mut_thres"],
                        fitnessval=np.array(mutation["fitnessval"]), fitnesstst=np.array(mutation["fitnesstst"]), complexity=np.array(mutation["complexity"]))

    GAparsimony._mutation(model)
    
    assert (model.population.population==np.array(mutation["resultado"])).all()

#################################################
#****************TEST CROSSOVER*****************#
#################################################

@pytest.mark.parametrize("crossover", [readJSONFile('./test/outputs/crossover.json')])
def test_GAParsimony_regression_boston_crossover(crossover):
    np.random.seed(crossover["seed"])
    
    model = GenericClass(pcrossover=crossover["pcrossover"],
                        popSize=crossover["popSize"], population=Population(crossover["params"], crossover["features"], np.array(crossover["population"])), 
                        fitnessval=np.array(crossover["fitnessval"]), 
                        fitnesstst=np.array(crossover["fitnesstst"]), complexity=np.array(crossover["complexity"]))

    nmating = int(np.floor(model.popSize/2))
    mating = np.random.choice(list(range(2 * nmating)), size=(2 * nmating), replace=False).reshape((nmating, 2))
    for i in range(nmating):
        if model.pcrossover > np.random.uniform(low=0, high=1):
            parents = mating[i, ]
            GAparsimony._crossover(model, parents=parents)
    
    assert (model.population.population==np.array(crossover["resultado"])).all()



# -*- coding: utf-8 -*-

import numpy as np
from GAparsimony.lhs import randomLHS_int
from GAparsimony.lhs.util import calculateSOptimal, calculateDistance, runifint, runif_std, findorder_zero

def geneticLHS(n, k, pop=100, gen=4, pMut=0.1, criterium="S", seed=None):
    r"""Latin Hypercube Sampling with a Genetic Algorithm

    Draws a Latin Hypercube Sample from a set of uniform distributions for use in creating 
    a LatinHypercube Design.  This function attempts to optimize the sample with respect 
    to the S optimalitycriterion through a genetic type algorithm.

    Parameters
    ----------
    n : int
        The number of rows or samples. 
    k : int
        The number of columns or parameters/variables.
    pop : int, optional
        The number of designs in the initial population. Default `100`.
    gen : int, optional
        The number of generations over which the algorithm is applied. Default `4`.
    pMut : float, optional
        The probability with which a mutation occurs in a column of the progeny. Default `0.1`.
    criterium : str, {'S', 'Maximin'}, optional
        The optimality criterium of the algorithm. Default is `'S'`. Maximin is also supported.
    seed : int, optional
        Random seed. Default `None`.

    Returns
    -------
    numpy.array
        A `numpy.array` of `float` with shape `(n, k)`.
    """

    if n < 1 or k < 1:
        raise Exception("nsamples are less than 1 (n) or nparameters less than 1 (k)")
    
    m_n = n
    m_k = k

    result = np.empty((m_n, m_k)).astype(np.double)
    if gen < 1 or pop < 1:
        raise Exception("pop, and gen should be integers greater than 0")
    
    m_pop = pop
    m_gen = gen
    
    if pMut <= 0 or pMut >= 1:
        raise Exception("pMut should be between 0 and 1")
    if m_pop % 2 != 0:
        raise Exception("pop should be an even number")

    if seed:
        np.random.seed(seed)

    A = np.empty(m_pop).astype(object)
    
    for i in range(m_pop):
        # // fill A with random hypercubes
        A[i] = randomLHS_int(m_n, m_k)
    
    for _ in range(m_gen):
        B = np.empty(m_pop).astype(np.double)
        for i in range(m_pop):
            if criterium == "S":
                B[i] = int(calculateSOptimal(A[i]))
            elif criterium == "Maximin":
                dist = calculateDistance(A[i]).astype(np.int32)[0]
                # // we want to find the minimum distance element, but there are zeros in the dist matrix
                distnonzero = []
                for mit in dist:
                    if mit > 0.0:
                        distnonzero.append(mit)

                distnonzero = np.array(distnonzero) 
                it = np.min(distnonzero) if distnonzero.shape[0]>0 else 0.0
                B[i] = it
            
            else: 
                raise Exception(f"Criterium not recognized: S and Maximin are available: {criterium} was provided.")

        # // H is used as an index on vector of matrices, A, so it should be using zero based order
        H = findorder_zero(B)
        posit = int(np.max(B) - B[0])
        J = np.empty(m_pop).astype(object)
        # // the first half of the next population gets the best hypercube from the first population
        for i in range(int(m_pop / 2)):
            J[i] = A[posit]
        
        if m_pop / 2 == 1:
            break
        
        # // the second half of the next population gets the decreasingly best hypercubes from the first population
        for i in range(int(m_pop / 2)):
            J[int(i + m_pop / 2)] = A[int(H[i])]
        
        # // skip the first best hypercube in the next generation
        # // in the others in the first half of the population, randomly permute a column from the second half into the first half
        for i in range(1, int(m_pop / 2)):
            temp1 = runifint(0, m_k-1).astype(np.int32)
            temp2 = runifint(0, m_k-1).astype(np.int32)
            for irow in range(m_n):
                J[i][irow, temp1] = J[int(i + m_pop / 2)][irow, temp2]
            
        # // for the second half of the population, randomly permute a column from the best hypercube
        for i in range(int(m_pop / 2), m_pop):
            temp1 = runifint(0, m_k-1).astype(np.int32)
            temp2 = runifint(0, m_k-1).astype(np.int32)
            for irow in range(m_n):
                J[i][irow, temp1] = A[posit][irow, temp2]
            
        # // randomly exchange two numbers in pMut percent of columns
        for i in range(1, m_pop):
            y = runif_std(m_k)
            for j in range(m_k):
                if y[j] <= pMut:
                    z = runifint(0, m_n-1, 2).astype(np.int32)
                    a = J[i][z[0], j]
                    b = J[i][z[1], j]
                    J[i][z[0], j] = b
                    J[i][z[1], j] = a
                
        # // put all of J back into A to start the next round
        A = J
    
    eps = runif_std(m_n * m_k)
    count = 0
    for j in range(m_k):
        for i in range(m_n):
            result[i,j] = (np.double(J[0][i,j]) - 1.0 + eps[count]) / np.double(m_n)
            count += 1
        
    return result





# -*- coding: utf-8 -*-

import inspect
import functools
import json
import numpy as np

def autoargs(*include, **kwargs):
    def _autoargs(func):
        attrs, varargs, _, defaults, _, _, _ = inspect.getfullargspec(func)

        def sieve(attr):
            if kwargs and attr in kwargs['exclude']:
                return False
            if not include or attr in include:
                return True
            else:
                return False

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            # handle default values
            if defaults:
                for attr, val in zip(reversed(attrs), reversed(defaults)):
                    if sieve(attr):
                        setattr(self, attr, val)
            # handle positional arguments
            positional_attrs = attrs[1:]
            for attr, val in zip(positional_attrs, args):
                if sieve(attr):
                    setattr(self, attr, val)
            # handle varargs
            if varargs:
                remaining_args = args[len(positional_attrs):]
                if sieve(varargs):
                    setattr(self, varargs, remaining_args)
            # handle varkw
            if kwargs:
                for attr, val in kwargs.items():
                    if sieve(attr):
                        setattr(self, attr, val)
            return func(self, *args, **kwargs)
        return wrapper
    return _autoargs

class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


def writeJSONFile(path, value):
    with open(path, 'w') as f:
        f.write(json.dumps(value, cls=NumpyEncoder))

def readJSONFile(path):
    with open(path) as f:
        return json.load(f)


# -*- coding: utf-8 -*-

import numpy  as np


# new function for monitoring
def parsimony_monitor(object, digits = 7, *args):
    r"""Functions for monitoring GA-PARSIMONY algorithm evolution

    Functions to print summary statistics of fitness values at each iteration of a GA search.

    Parameters
    ----------
    object : object of GAparsimony
        The `GAparsimony` object that we want to monitor .
    digits : int
        Minimal number of significant digits.
    *args : 
        Further arguments passed to or from other methods.
    """

    fitnessval = object.fitnessval[~np.isnan(object.fitnessval)]
        
    print(f"GA-PARSIMONY | iter = {object.iter}")
    print("|".join([f"MeanVal = {round(np.mean(fitnessval), digits)}".center(16+digits), 
                    f"ValBest = {round(object.bestfitnessVal, digits)}".center(16+digits),
                    f"TstBest = {round(object.bestfitnessTst, digits)}".center(16+digits),
                    f"ComplexBest = {round(object.bestcomplexity, digits)}".center(19+digits),
                    f"Time(min) = {round(object.minutes_gen, digits)}".center(17+digits)]) + "\n")
 



# Duda si es todo el rato con x1
# Equivalencia a fivenum es np.percentile(aux, [0, 25, 50, 75, 100])

def parsimony_summary(object, *args):
  
    x1 = object.fitnessval[~np.isnan(object.fitnessval)]
    q1 = np.percentile(x1, [0, 25, 50, 75, 100])
    x2 = object.fitnesstst[~np.isnan(object.fitnesstst)]
    q2 = np.percentile(x1, [0, 25, 50, 75, 100])
    x3 = object.complexity[~np.isnan(object.complexity)]
    q3 = np.percentile(x1, [0, 25, 50, 75, 100])

    return q1[4], np.mean(x1), q1[3], q1[2], q1[1], q1[0], q2[4], np.mean(x2), q2[3], q2[2], q2[1], q2[0], q3[4], np.mean(x3), q3[3], q3[2], q3[1], q3[0]






# -*- coding: utf-8 -*-

import numpy as np
from .bclib import inner_product

def isValidLHS_int(matrix):
    for jcol in range(matrix.shape[1]):
        total = 0
        for irow in range(matrix.shape[0]):
            total = total + matrix[irow, jcol]
        if not total == int(matrix.shape[0] * (matrix.shape[0] + 1) / 2):
        
            return False
        
    return True

def isValidLHS(matrix):

    n = matrix.shape[0]
    k = matrix.shape[1]
    resultint = np.empty((n, k))
    # for (;it != result.end(); ++it, ++iti) # Itera sobre filas, fichero matrix.h
    for i in range(matrix.shape[0]):
        resultint[i] = 1 + (np.floor(np.double(n) * (matrix[i]))).astype(np.int32)

    for jcol in range(resultint.shape[1]):
        total = 0
        for irow in range(resultint.shape[0]):
            total = total + resultint[irow, jcol]
        if not total == int(resultint.shape[0] * (resultint.shape[0] + 1) / 2):
        
            return False
        
    return True     

def initializeAvailableMatrix(i, j):
    dev = np.empty((i, j)).astype(np.double)
    for irow in range(i):
    
        for jcol in range(j):
        
            dev[irow, jcol] = np.double(jcol + 1)
        
    return dev

def runif_std(n):
    dev = np.empty(n).astype(np.double)
    for i in range(n):
        dev[i] =np.random.uniform(low=0, high=1)
    return dev

def convertIntegerToNumericLhs(intMat):
  n = intMat.shape[0]
  k = intMat.shape[1]
  result = np.empty((n, k))
  eps = np.random.rand(n*k)
  counter = 0
  # // I think this is right (iterate over rows within columns
  for col in range(k):
      for row in range(n):
        result[row, col] = np.double(intMat[row, col] - 1) + eps[counter]
        result[row, col] = result[row, col] / np.double(n)
        counter+=1
  return result


def sumInvDistance(a): 
	return np.sum(calculateDistance(a)[::-1]) # equals to accumulate
    

def calculateDistanceSquared(a, b):

    if a.shape != b.shape:
        raise Exception("Inputs of a different size")
    
    return inner_product(a, b, np.double(0), lambda a, b: a+b, lambda x, y: (x-y) * (x-y))

def calculateDistance(mat):
    m_rows = mat.shape[0]
    result = np.empty((m_rows, m_rows)).astype(np.double)
    for i in range(m_rows - 1):
        for j in range(i+1, m_rows):
            result[i,j] = np.sqrt(np.double(calculateDistanceSquared(mat[i,:], mat[j,:])))
    return result
        
    
def calculateSOptimal(mat):
    return 1.0 / sumInvDistance(mat)


def runifint(a, b, n=None):
    if not n:
        return  a + (np.floor((np.random.uniform(low=0, high=1) * (np.double(b) + 1.0 - np.double(a)))))
    else:
        result = np.empty(n).astype(np.double)
        r = runif_std(n)
        for i in range(n):
            result[i] = a + (np.floor(r[i] * (b + 1.0 - b + 1.0 - a)))
        return result

    


# -*- coding: utf-8 -*-

from GAparsimony.lhs.util import isValidLHS, isValidLHS_int
from GAparsimony.lhs import geneticLHS, improvedLHS, maximinLHS, optimumLHS, randomLHS, randomLHS_int

import pytest

@pytest.mark.parametrize("shape", [
    (2, 2),
    (6, 6),
    (3, 8)
])
def test_randomLHS_int(shape):
    assert isValidLHS_int(randomLHS_int(*shape))

@pytest.mark.parametrize("shape", [
    (2, 2),
    (6, 6),
    (3, 8)
])
def test_randomLHS(shape):
    assert isValidLHS(randomLHS(*shape))

@pytest.mark.parametrize("shape", [
    (2, 2),
    (6, 6),
    (3, 8)
])
def test_geneticLHS(shape):
    assert isValidLHS(geneticLHS(*shape))

@pytest.mark.parametrize("shape", [
    (2, 2),
    (6, 6),
    (3, 8)
])
def test_improvedLHS(shape):
    assert isValidLHS(improvedLHS(*shape))

@pytest.mark.parametrize("shape", [
    (2, 2),
    (6, 6),
    (3, 8)
])
def test_maximinLHS(shape):
    assert isValidLHS(maximinLHS(*shape))

@pytest.mark.parametrize("shape", [
    (2, 2),
    (6, 6),
    (8, 8)
])
def test_optimumLHS(shape):
    assert isValidLHS(optimumLHS(*shape))


# -*- coding: utf-8 -*-

import numpy as np

# Ya estaría implementadas

def findorder_zero(v):
    
    # create a vector of pairs to hold the value and the integer rank
    p = np.empty(v.shape, dtype=object)
    for i, vi in enumerate(v):
        p[i] = (vi, i)
    p = np.array(list(sorted(p, key=lambda a: a[0])))
    
    return np.array(list(map(lambda x: x[1], p)))


def findorder(v):
    order = findorder_zero(v)
    for i in range(len(order)):
        order[i] += 1
    return order
 

def inner_product(a, b, init=0, op1=lambda a, b: a + b, op2=lambda a, b: a * b):
    for x, y in zip(a, b):
        init = op1(init, op2(x, y))
    return init



# -*- coding: utf-8 -*-

from sklearn.model_selection import train_test_split, cross_val_score, RepeatedKFold
from sklearn.metrics import make_scorer
import numpy as np

def getFitness(algorithm, metric, complexity, cv=RepeatedKFold(n_splits=10, n_repeats=5, random_state=42), minimize=False, test_size=0.2, random_state=42, n_jobs=-1):
    r"""
     Fitness function for GAparsimony.

    Parameters
    ----------
    algorithm : object
        The machine learning function to optimize. 
    metric : function
        A function that computes the fitness value.
    complexity : function
        A function that calculates the complexity of the model. There are some functions available in `GAparsimony.util.complexity`.
    cv : object, optional
        An `sklearn.model_selection` function. By default, is defined `RepeatedKFold(n_splits=10, n_repeats=5, random_state=42)`.
    minimize : bool, optional
        `False`, if the objective is to minimize the metric, to maximize it, set to `True`.
    test_size : float, int, None
        If float, should be between 0.0 and 1.0 and represent the proportion
        of the dataset to include in the test split. If int, represents the
        absolute number of test samples. If None, the value is set to the
        complement of the train size. Default 0.2.
    random_state : int, optional
        Controls the shuffling applied to the data before applying the split.
        Pass an int for reproducible output across multiple function calls.
        Default `42`
    n_jobs : int, optional
        Number of jobs to run in parallel. Training the estimator and computing
        the score are parallelized over the cross-validation splits.
        ``-1`` means using all processors. Default `-1`
    Examples
    --------
    Usage example for a regression model 
    
    .. highlight:: python
    .. code-block:: python

        from sklearn.svm import SVC
        from sklearn.metrics import cohen_kappa_score

        from GAparsimony import getFitness
        from GAparsimony.util import svm_complexity

        fitness = getFitness(SVC, cohen_kappa_score, svm_complexity, cv, maximize=True, test_size=0.2, random_state=42, n_jobs=-1)
    """

    if algorithm is None:
        raise Exception("An algorithm function must be provided!!!")
    if metric is None or not callable(metric):
        raise Exception("A metric function must be provided!!!")
    if complexity is None or not callable(complexity):
        raise Exception("A complexity function must be provided!!!")
    
    def fitness(cromosoma, **kwargs):
        if "pandas" in str(type(kwargs["X"])):
            kwargs["X"] = kwargs["X"].values
        if "pandas" in str(type(kwargs["y"])):
            kwargs["y"] = kwargs["y"].values
        X_train, X_test, y_train, y_test = train_test_split(kwargs["X"], kwargs["y"], test_size=test_size, random_state=random_state)
        try:
            # Extract features from the original DB plus response (last column)
            data_train_model = X_train[: , cromosoma.columns] 
            data_test_model = X_test[: , cromosoma.columns] 

            # train the model

            aux = algorithm(**cromosoma.params)
            fitness_val = cross_val_score(aux, data_train_model, y_train, scoring=make_scorer(metric), cv=cv, n_jobs=n_jobs).mean()
            modelo = algorithm(**cromosoma.params).fit(data_train_model, y_train)
            fitness_test = metric(modelo.predict(data_test_model), y_test)

            if minimize:
                fitness_val = -fitness_val
                fitness_test = -fitness_test

            return np.array([fitness_val, fitness_test, complexity(modelo, np.sum(cromosoma.columns))]), modelo
        except Exception as e:    
            print(e)
            return np.array([np.NINF, np.NINF, np.Inf]), None

    return fitness



# -*- coding: utf-8 -*-

"""Combines feature selection(FS), hyperparameter tuning (HT), and parsimonious model selection (PMS) 
with Genetic Algorithm (GA) optimization. GA selection procedure is based on separate cost and complexity 
evaluations. Therefore, the best individuals are initially sorted by an error fitness function, and afterwards, 
models with similar costs are rearranged according to the model complexity measurement so as to foster models 
of lesser complexity. The algorithm can be run sequentially or in parallel.

GAparsimonypackage is a new GA wrapper automatic method that efficiently generated machine learning  
models with reduced complexity and adequate generalization capacity.ga_parsimonyfunction is primarily based on 
combining FS and HT with a second novel GA selection process (named ReRank algorithm) 
in order to achieve better overall parsimonious models. Unlike other GA methodologies that use a penalty parameter 
for combining loss and complexity measures into a unique fitness function, the main contribution of this package 
is that ga_parsimony selects the best models by considering cost and complexity separately. For this purpose, the 
*ReRank* algorithm rearranges individuals by their complexity when there is not a significant difference between 
their costs. Thus, less complex models with similar accuracy are promoted. Furthermore, because the penalty 
parameter is unnecessary, there is no consequent uncertainty associated with assigning a correct value beforehand. 
As a result, with GAPARSIMONY, an automatic method for obtaining parsimonious models is finally made possible.

References
----------
F.J. Martinez-de-Pison, J. Ferreiro, E. Fraile, A. Pernia-Espinoza, A comparative study of six model complexity 
metrics to search for parsimonious models with GAparsimony R Package, Neurocomputing,
Volume 452, 2021, Pages 317-332, ISSN 0925-2312, https://doi.org/10.1016/j.neucom.2020.02.135.

Martinez-de-Pison, F.J., Gonzalez-Sendino, R., Aldama, A., Ferreiro-Cabello, J., Fraile-Garcia, E. Hybrid methodology 
based on Bayesian optimization and GA-PARSIMONY to search for parsimony models by combining hyperparameter optimization 
and feature selection (2019) Neurocomputing, 354, pp. 20-26. https://doi.org/10.1016/j.neucom.2018.05.136

Urraca R., Sodupe-Ortega E., Antonanzas E., Antonanzas-Torres F., Martinez-de-Pison, F.J. (2017). Evaluation of a 
novel GA-based methodology for model structure selection: The GA-PARSIMONY. Neurocomputing, Online July 2017. https://doi.org/10.1016/j.neucom.2016.08.154

Martinez-De-Pison, F.J., Gonzalez-Sendino, R., Ferreiro, J., Fraile, E., Pernia-Espinoza, A. GAparsimony: An R 
package for searching parsimonious models by combining hyperparameter optimization and feature selection (2018) Lecture 
Notes in Computer Science (including subseries Lecture Notes in Artificial Intelligence and Lecture Notes in Bioinformatics), 
10870 LNAI, pp. 62-73. https://doi.org/10.1007/978-3-319-92639-1_6

Applications
------------
Eduardo Dulce-Chamorro, Francisco Javier Martinez-de-Pison, An advanced methodology to enhance energy efficiency in 
a hospital cooling-water system, Journal of Building Engineering, Volume 43,
2021, 102839, ISSN 2352-7102,https://doi.org/10.1016/j.jobe.2021.102839.

Sanz-Garcia, A., Fernandez-Ceniceros, J., Antonanzas-Torres, F., Pernia-Espinoza, A.V., Martinez-De-Pison, 
F.J. GA-PARSIMONY: A GA-SVR approach with feature selection and parameter optimization to obtain parsimonious solutions 
for predicting temperature settings in a continuous annealing furnace (2015) Applied Soft Computing Journal, 35, art. 
no. 3006, pp. 13-28. https://doi.org/10.1016/j.asoc.2015.06.012

Fernandez-Ceniceros, J., Sanz-Garcia, A., Antoñanzas-Torres, F., Martinez-de-Pison, F.J. A numerical-informational 
approach for characterising the ductile behaviour of the T-stub component. Part 2: Parsimonious soft-computing-based 
metamodel (2015) Engineering Structures, 82, pp. 249-260. https://doi.org/10.1016/j.engstruct.2014.06.047

Antonanzas-Torres, F., Urraca, R., Antonanzas, J., Fernandez-Ceniceros, J., Martinez-De-Pison, F.J. Generation of 
daily global solar irradiation with support vector machines for regression (2015) Energy Conversion and Management, 
96, pp. 277-286. https://doi.org/10.1016/j.enconman.2015.02.086


"""

from GAparsimony import Population, order
from GAparsimony.util import parsimony_monitor, parsimony_summary
from GAparsimony.lhs import geneticLHS, improvedLHS, maximinLHS, optimumLHS, randomLHS

import warnings
import numpy as np
import pandas as pd
pd.set_option('display.max_columns', None)

import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib.lines as mlines
import matplotlib.patches as mpatches
import time
import inspect, opcode




class GAparsimony(object):

    def __init__(self, 
                fitness,
                params,
                features,
                type_ini_pop="improvedLHS", 
                popSize = 50, 
                pcrossover = 0.8,  
                maxiter = 40, 
                feat_thres=0.90, 
                rerank_error = 0.0, 
                iter_start_rerank = 0,
                pmutation = 0.10, 
                feat_mut_thres=0.10, 
                not_muted=3,
                elitism = None,
                selection = "nlinear", 
                keep_history = False,
                early_stop = None, 
                maxFitness = np.Inf, 
                suggestions = None,
                seed_ini = None, 
                verbose=1):
        r"""
        A class for searching parsimonious models by feature selection and parameter tuning with
        genetic algorithms.

        Parameters
        ----------
        fitness : function
            The fitness function, any function which takes as input a chromosome which combines the model parameters 
            to tune and the features to be selected. Fitness function returns a numerical vector with three values:validation_cost, 
            testing_cost and model_complexity, and the trained model. 
        params : dict
            It is a dictionary with the model's hyperparameters to be adjusted and the range of values to search for.

            .. code-block::

                {
                    "<< hyperparameter name >>": 
                    {
                        "range": [<< minimum value >>, << maximum value >>],
                        "type": GAparsimony.FLOAT/GAparsimony.INTEGER
                    },
                    "<< hyperparameter name >>": 
                    {
                        "value": << constant value >>,
                        "type": GAparsimony.CONSTANT
                    }
                }

        features : int or list of str
            The number of features/columns in the dataset or a list with their names.
        type_ini_pop : str, {'randomLHS', 'geneticLHS', 'improvedLHS', 'maximinLHS', 'optimumLHS', 'random'}, optional
            Method to create the first population with `GAparsimony._population` function. Possible values: `randomLHS`, `geneticLHS`, 
            `improvedLHS`, `maximinLHS`, `optimumLHS`, `random`.First 5 methods correspond with several latine hypercube for initial sampling. By default is set to `improvedLHS`.
        popSize : int, optional
            The population size.
        pcrossover : float, optional
            The probability of crossover between pairs of chromosomes. Typically this is alarge value and by default is set to `0.8`.
        maxiter : float, optional
            The maximum number of iterations to run before the GA process is halted.
        feat_thres : float, optional
            Proportion of selected features in the initial population. It is recommended a high percentage of the selected features for 
            the first generations. By default is set to `0.90`.
        rerank_error : float, optional
            When a value is provided, a second reranking process according to the model complexities is called by `parsimony_rerank` function. 
            Its primary objective isto select individuals with high validation cost while maintaining the robustnessof a parsimonious model. 
            This function switches the position of two models if the first one is more complex than the latter and no significant difference 
            is found between their fitness values in terms of cost. Thus, if the absolute difference between the validation costs are 
            lower than `rerank_error` they areconsidered similar. Default value=`0.01`.
        iter_start_rerank : int, optional
            Iteration when ReRanking process is actived. Default=`0`. Sometimes is useful not to use ReRanking process in the first generations.
        pmutation : float, optional
            The probability of mutation in a parent chromosome. Usually mutation occurswith a small probability. By default is set to `0.10`.
        feat_mut_thres : float, optional
            Probability of the muted `features-chromosome` to be one. Default value is set to `0.10`.
        not_muted : int, optional
            Number of the best elitists that are not muted in each generation. Default valueis set to `3`.
        elitism : int, optional
            The number of best individuals to survive at each generation. By default the top `20%` individuals will survive at each iteration.
        selection : str, optional
            Method to perform selection with `GAparsimony._selection` function. Possible values: `linear`, `nlinear`, `random`. By default is set to `nlinear`.
        keep_history : bool, optional
            If it is `True` keeps in the list GAparsimony.history each generation as `pandas.DataFrame`. This parameter must set `True` in 
            order to use `GAparsimony.plot` method or `GAparsimony.importance` function.
        maxFitness : int, optional
            The upper bound on the fitness function after that the GA search is interrupted. Default value is set to +Inf.
        early_stop : int, optional
            The number of consecutive generations without any improvement in the bestfitness value before the GA is stopped.
        suggestions : numpy.array, optional
            A matrix of solutions strings to be included in the initial population.
        seed_ini : int, optional
            An integer value containing the random number generator state.
        verbose : int, optional
            The level of messages that we want it to show us. Possible values: 1=monitor level,  2=debug level, if 0 no messages. Default `1`.


        Attributes
        ----------
        population : Population
            The current (or final) population.
        minutes_total : float
            Total elapsed time (in minutes).
        history : float
            A list with the population of all iterations.
        best_score : float
            The best validation score in the whole GA process.
        best_model
            The best model in the whole GA process.
        best_model_conf : Chromosome
            The parameters and features of the best model in the whole GA process.
        bestfitnessVal : float
            The validation cost of the best solution at the last iteration.
        bestfitnessTst : float
            The testing cost of the best solution at the last iteration.
        bestcomplexity : float
            The model complexity of the best solution at the last iteration.

        Examples
        --------
        Usage example for a regression model using the sklearn boston dataset 

        .. highlight:: python
        .. code-block:: python

            from sklearn.model_selection import RepeatedKFold
            from sklearn.linear_model import Lasso
            from sklearn.preprocessing import StandardScaler
            from sklearn.metrics import mean_squared_error

            from sklearn.datasets import load_boston

            from GAparsimony import GAparsimony, Population, getFitness
            from GAparsimony.util import linearModels_complexity

            boston = load_boston()
            X, y = boston.data, boston.target 
            X = StandardScaler().fit_transform(X)

            # ga_parsimony can be executed with a different set of 'rerank_error' values
            rerank_error = 0.01

            params = {"alpha":{"range": (1., 25.9), "type": Population.FLOAT}, 
                        "tol":{"range": (0.0001,0.9999), "type": Population.FLOAT}}

            cv = RepeatedKFold(n_splits=10, n_repeats=10, random_state=42)

            fitness = getFitness(Lasso, mean_squared_error, linearModels_complexity, cv, minimize=True, 
                                 test_size=0.2, random_state=42, n_jobs=-1)


            GAparsimony_model = GAparsimony(fitness=fitness,
                                        params = params, 
                                        features = boston.feature_names,
                                        keep_history = True,
                                        rerank_error = rerank_error,
                                        popSize = 40,
                                        maxiter = 50, early_stop=10,
                                        feat_thres=0.90, # Perc selected features in first generation
                                        feat_mut_thres=0.10, # Prob of a feature to be one in mutation
                                        seed_ini = 1234)


            GAparsimony_model.fit(X, y)

            GAparsimony_model.summary()

            aux = GAparsimony_model.summary()

            GAparsimony_model.plot()

        .. code-block:: text

            GA-PARSIMONY | iter = 0
            MeanVal = -79.1813338 | ValBest = -30.3470614 | TstBest = -29.2466835 |ComplexBest = 13000000021.927263|  Time(min) = 0.185549

            GA-PARSIMONY | iter = 1
            MeanVal = -55.0713465 | ValBest = -30.2283235 | TstBest = -29.2267507 |ComplexBest = 12000000022.088743| Time(min) = 0.1238126  

            GA-PARSIMONY | iter = 2
            MeanVal = -34.8473723 | ValBest = -30.2283235 | TstBest = -29.2267507 |ComplexBest = 12000000022.088743| Time(min) = 0.0907046  

            GA-PARSIMONY | iter = 3
            MeanVal = -38.5251529 | ValBest = -30.0455259 | TstBest = -29.2712578 |ComplexBest = 10000000022.752678| Time(min) = 0.0755356  

            ...

            GA-PARSIMONY | iter = 20
            MeanVal = -34.2636095 | ValBest = -29.5036901 | TstBest = -29.3245069 |ComplexBest = 5000000023.115818| Time(min) = 0.0659549  

            GA-PARSIMONY | iter = 21
            MeanVal = -40.4629864 | ValBest = -29.5036901 | TstBest = -29.3245069 |ComplexBest = 5000000023.115818| Time(min) = 0.0725066  

            GA-PARSIMONY | iter = 22
            MeanVal = -35.9230384 | ValBest = -29.5036901 | TstBest = -29.3245069 |ComplexBest = 5000000023.115818| Time(min) = 0.0704362  

            GA-PARSIMONY | iter = 23
            MeanVal = -36.5946762 | ValBest = -29.5036901 | TstBest = -29.3245069 |ComplexBest = 5000000023.115818| Time(min) = 0.0723252  

            GA-PARSIMONY | iter = 24
            MeanVal = -37.3293511 | ValBest = -29.5036901 | TstBest = -29.3245069 |ComplexBest = 5000000023.115818| Time(min) = 0.0684883  

            +------------------------------------+
            |             GA-PARSIMONY           |
            +------------------------------------+

            GA-PARSIMONY settings:
            Number of Parameters      = 2
            Number of Features        = 13
            Population size           = 40
            Maximum of generations    = 50
            Number of early-stop gen. = 10
            Elitism                   = 8
            Crossover probability     = 0.8
            Mutation probability      = 0.1
            Max diff(error) to ReRank = 0.01
            Perc. of 1s in first popu.= 0.9
            Prob. to be 1 in mutation = 0.1

            Search domain =
                    alpha     tol  CRIM   ZN  INDUS  CHAS  NOX   RM  AGE  DIS  RAD  \
            Min_param    1.0  0.0001   0.0  0.0    0.0   0.0  0.0  0.0  0.0  0.0  0.0
            Max_param   25.9  0.9999   1.0  1.0    1.0   1.0  1.0  1.0  1.0  1.0  1.0

                    TAX  PTRATIO    B  LSTAT
            Min_param  0.0      0.0  0.0    0.0
            Max_param  1.0      1.0  1.0    1.0


            GA-PARSIMONY results:
            Iterations                = 25
            Best validation score = -29.502012171608403


            Solution with the best validation score in the whole GA process =

            fitnessVal fitnessTst complexity    alpha       tol CRIM ZN INDUS CHAS NOX  \
            0    -29.502   -29.3244      5e+09  1.33694  0.541197    0  0     0    0   0

            RM AGE DIS RAD TAX PTRATIO  B LSTAT
            0  1   1   0   0   0       1  1     1


            Results of the best individual at the last generation =

            Best indiv's validat.cost = -29.503690126221098
            Best indiv's testing cost = -29.324506895493244
            Best indiv's complexity   = 5000000023.115818
            Elapsed time in minutes   = 2.031593410174052


            BEST SOLUTION =

            fitnessVal fitnessTst complexity   alpha       tol CRIM ZN INDUS CHAS NOX  \
            0   -29.5037   -29.3245      5e+09  1.3374  0.547189    0  0     0    0   0

            RM AGE DIS RAD TAX PTRATIO  B LSTAT
            0  1   1   0   0   0       1  1     1
        
        .. figure:: ../docs/img/regression.png
            :align: center
            :width: 600
            :alt: Regression plot

            Regression plot
        
        Usage example for a classification model using the wine dataset 

        .. highlight:: python
        .. code-block:: python

            from sklearn.model_selection import RepeatedKFold
            from sklearn.svm import SVC
            from sklearn.metrics import cohen_kappa_score
            from sklearn.datasets import load_wine

            from GAparsimony import GAparsimony, Population, getFitness
            from GAparsimony.util import svm_complexity

            wine = load_wine()
            X, y = wine.data, wine.target 
            X = StandardScaler().fit_transform(X)

            rerank_error = 0.001
            params = {"C":{"range": (00.0001, 99.9999), "type": Population.FLOAT}, 
                    "gamma":{"range": (0.00001,0.99999), "type": Population.FLOAT}, 
                    "kernel": {"value": "poly", "type": Population.CONSTANT}}

            cv = RepeatedKFold(n_splits=10, n_repeats=10, random_state=42)

            fitness = getFitness(SVC, cohen_kappa_score, svm_complexity, cv, minimize=False, test_size=0.2, random_state=42, n_jobs=-1)

            GAparsimony_model = GAparsimony(fitness=fitness,
                                            params=params,
                                            features=wine.feature_names,
                                            keep_history = True,
                                            rerank_error = rerank_error,
                                            popSize = 40,
                                            maxiter = 50, early_stop=10,
                                            feat_thres=0.90, # Perc selected features in first generation
                                            feat_mut_thres=0.10, # Prob of a feature to be one in mutation
                                            seed_ini = 1234)



            GAparsimony_model.fit(X, y)

            GAparsimony_model.summary()

            GAparsimony_model.plot()

        .. code-block:: text

            GA-PARSIMONY | iter = 0
            MeanVal = 0.879549  |  ValBest = 0.9314718  |  TstBest = 0.9574468  |ComplexBest = 10000000045.0| Time(min) = 0.1438692  

            GA-PARSIMONY | iter = 1
            MeanVal = 0.9075035  |  ValBest = 0.9496819  |  TstBest = 0.9142857  |ComplexBest = 11000000060.0| Time(min) = 0.0893566  

            GA-PARSIMONY | iter = 2
            MeanVal = 0.9183232  |  ValBest = 0.9496819  |  TstBest = 0.9142857  |ComplexBest = 11000000060.0| Time(min) = 0.0818844  

            GA-PARSIMONY | iter = 3
            MeanVal = 0.9219764  |  ValBest = 0.9534295  |  TstBest = 0.9568345  |ComplexBest = 10000000043.0| Time(min) = 0.0739248 

            ...

            GA-PARSIMONY | iter = 19
            MeanVal = 0.9182586  |   ValBest = 0.972731  |     TstBest = 1.0     |ComplexBest = 7000000048.0| Time(min) = 0.0867344  

            GA-PARSIMONY | iter = 20
            MeanVal = 0.9224294  |   ValBest = 0.972731  |     TstBest = 1.0     |ComplexBest = 7000000048.0| Time(min) = 0.0771279  

            GA-PARSIMONY | iter = 21
            MeanVal = 0.9150223  |   ValBest = 0.972731  |     TstBest = 1.0     |ComplexBest = 7000000048.0| Time(min) = 0.0847196  

            GA-PARSIMONY | iter = 22
            MeanVal = 0.9335024  |   ValBest = 0.972731  |     TstBest = 1.0     |ComplexBest = 7000000048.0| Time(min) = 0.0814945 


            +------------------------------------+
            |             GA-PARSIMONY           |
            +------------------------------------+

            GA-PARSIMONY settings:
            Number of Parameters      = 2
            Number of Features        = 13
            Population size           = 40
            Maximum of generations    = 50
            Number of early-stop gen. = 10
            Elitism                   = 8
            Crossover probability     = 0.8
            Mutation probability      = 0.1
            Max diff(error) to ReRank = 0.001
            Perc. of 1s in first popu.= 0.9
            Prob. to be 1 in mutation = 0.1

            Search domain =
                            C    gamma  alcohol  malic_acid  ash  alcalinity_of_ash  \
            Min_param   0.0001  0.00001      0.0         0.0  0.0                0.0
            Max_param  99.9999  0.99999      1.0         1.0  1.0                1.0

                    magnesium  total_phenols  flavanoids  nonflavanoid_phenols  \
            Min_param        0.0            0.0         0.0                   0.0
            Max_param        1.0            1.0         1.0                   1.0

                    proanthocyanins  color_intensity  hue  \
            Min_param              0.0              0.0  0.0
            Max_param              1.0              1.0  1.0

                    od280/od315_of_diluted_wines  proline
            Min_param                           0.0      0.0
            Max_param                           1.0      1.0


            GA-PARSIMONY results:
            Iterations                = 23
            Best validation score = 0.9727309855126027


            Solution with the best validation score in the whole GA process =

            fitnessVal fitnessTst complexity        C      gamma alcohol malic_acid ash  \
            0   0.972731          1      7e+09  51.1573  0.0581044       1          0   1

            alcalinity_of_ash magnesium total_phenols flavanoids nonflavanoid_phenols  \
            0                 1         0             0          1                    0

            proanthocyanins color_intensity hue od280/od315_of_diluted_wines proline
            0               0               0   1                            1       1


            Results of the best individual at the last generation =

            Best indiv's validat.cost = 0.9727309855126027
            Best indiv's testing cost = 1.0
            Best indiv's complexity   = 7000000048.0
            Elapsed time in minutes   = 1.9634766817092892


            BEST SOLUTION =

            fitnessVal fitnessTst complexity        C      gamma alcohol malic_acid ash  \
            0   0.972731          1      7e+09  51.1573  0.0581044       1          0   1

            alcalinity_of_ash magnesium total_phenols flavanoids nonflavanoid_phenols  \
            0                 1         0             0          1                    0

            proanthocyanins color_intensity hue od280/od315_of_diluted_wines proline
            0               0               0   1                            1       1
        
        .. figure:: ../docs/img/classification.png
            :align: center
            :width: 600
            :alt: Classification plot

            Classification plot
        """
        
        self.elitism = max(1, round(popSize * 0.20)) if not elitism else elitism

        self.population = Population(params, columns=features)


        # Check parameters
        # ----------------
        if selection is None or type(selection) is not str:
            raise ValueError("A selection(sting) must be provided!!!")
        if fitness is None:
            raise Exception("A fitness function must be provided!!!")
        if not callable(fitness):
            raise Exception("A fitness function must be provided!!!")
        if popSize < 10:
            warnings.warn("The population size is less than 10!!!")
        if maxiter < 1:
            raise ValueError("The maximum number of iterations must be at least 1!!!")
        if self.elitism > popSize:
            raise ValueError("The elitism cannot be larger that population size.")
        if pcrossover < 0 or pcrossover > 1:
            raise ValueError("Probability of crossover must be between 0 and 1!!!")
        if pmutation < 0 or pmutation > 1:
            raise ValueError("Probability of mutation must be between 0 and 1!!!")
        if self.population._min is None and self.population._max is None:
            raise ValueError("A min and max range of values must be provided!!!")
        if self.population._min.shape != self.population._max.shape:
            raise Exception("min_param and max_param must have the same length!!!")
        if features is None:
            raise Exception("Number of features or name of features must be provided!!!")
        if (suggestions is not None) or (type(suggestions) is list and len(suggestions)>0 and type(suggestions[0]) is not list) or (type(suggestions) is np.array and len(suggestions.shape) < 2):
            raise Exception("Provided suggestions is a vector")
        if (type(suggestions) is np.array) and (self.population._min.shape + features) != suggestions.shape[1]:
            raise Exception("Provided suggestions (ncol) matrix do not match the number of variables (model parameters + vector with selected features) in the problem!")

        self.fitness = fitness
        self.popSize = popSize
        self.pcrossover = pcrossover
        self.maxiter = maxiter
        self.feat_thres=feat_thres
        self.rerank_error=rerank_error
        self.iter_start_rerank=iter_start_rerank
        self.pmutation = pmutation
        self.feat_mut_thres=feat_mut_thres
        self.not_muted=not_muted
        
        self.selection = selection
        self.keep_history = keep_history
        self.early_stop = maxiter if not early_stop else early_stop
        self.maxFitness = maxFitness
        self.suggestions = np.array(suggestions) if type(suggestions) is list else suggestions
        self.seed_ini = seed_ini
        self.verbose = verbose

        self.iter = 0
        self.minutes_total=0
        self.best_score = np.NINF
        self.history = list()

        if self.seed_ini:
            np.random.seed(self.seed_ini)

        self.population.population = self._population(type_ini_pop=type_ini_pop) # Creo la poblacion de la primera generacion

        if self.suggestions:
            ng = min(self.suggestions.shape[0], popSize)
            if ng > 0:
                self.population[:ng, :] = self.suggestions[:ng, :]
    
    def fit(self, X, y, iter_ini=0):
        r"""
        A GA-based optimization method for searching accurate parsimonious models by combining feature selection, model tuning, 
        and parsimonious model selection (PMS). PMS procedure is basedon separate cost and complexity evaluations. The best individuals 
        are initially sorted by an *errorfitness* function, and afterwards, models with similar costs are rearranged according to their 
        modelcomplexity so as to foster models of lesser complexity.

        Parameters
        ----------
        X : pandas.DataFrame or numpy.array
            Training vector.
        y : pandas.DataFrame or numpy.array
            Target vector relative to X.
        iter_ini : int, optional
            Iteration/generation of `GAparsimony.history` to be used when model is pretrained. If
            `iter_ini==None` uses the last iteration of the model.

        """
           
        # Get suggestions
        # ---------------
        if self.verbose == 2 and self.suggestions:
            print(self.suggestions)


        # Initial settings
        # ----------------
        if self.seed_ini:
            np.random.seed(self.seed_ini)

        self._summary = np.empty((self.maxiter,6*3,))
        self._summary[:] = np.nan
        self.bestSolList = list()
        self.fitnessval = np.empty(self.popSize)
        self.fitnessval[:] = np.nan
        self.fitnesstst = np.empty(self.popSize)
        self.fitnesstst[:] = np.nan
        self.complexity = np.empty(self.popSize)
        self.complexity[:] = np.nan
        _models = np.empty(self.popSize).astype(object)


        if len(self.history) > 0:
            if self.verbose == 2:
                print("There is a GAparsimony 'object'!!!")
                print(self)

            iter_ini = self.iter if not iter_ini else min(iter_ini, self.iter)
            if iter_ini < 0:
                iter_ini = 0

            self.history = self.history[iter_ini].values[0]
            if self.verbose == 2:
                print(f"Starting GA optimization with a provided GAparsimony 'object'. Using object's GA settings and its population from iter={iter_ini}.")
        
        elif self.verbose == 2:
            print("\nStep 0. Initial population")
            print(np.c_[self.fitnessval, self.fitnesstst, self.complexity, self.population.population][:10, :])
            # input("Press [enter] to continue")


        # Main Loop
        # --------- 
        for iter in range(self.maxiter):
            tic = time.time()
            
            self.iter = iter

            for t in range(self.popSize):
                c = self.population.getChromosome(t)
                if np.isnan(self.fitnessval[t]) and np.sum(c.columns)>0:
                    fit = self.fitness(c, X=X, y=y)
                    self.fitnessval[t] = fit[0][0]
                    self.fitnesstst[t] = fit[0][1]
                    self.complexity[t] = fit[0][2]
                    _models[t] = fit[1]
                

            if self.seed_ini:
                np.random.seed(self.seed_ini*iter) 
            

            # Sort by the Fitness Value
            # ----------------------------
            sort = order(self.fitnessval, kind='heapsort', decreasing = True, na_last = True)
            self.population.population = self.population[sort, :]
            self.fitnessval = self.fitnessval[sort]
            self.fitnesstst = self.fitnesstst[sort]
            self.complexity = self.complexity[sort]
            _models = _models[sort]

            PopSorted = self.population._pop.copy()
            FitnessValSorted = self.fitnessval.copy()
            FitnessTstSorted = self.fitnesstst.copy()
            ComplexitySorted = self.complexity.copy()
            
            if np.max(self.fitnessval)>self.best_score:
                self.best_score = np.nanmax(self.fitnessval)
                self.solution_best_score = np.r_[self.best_score, 
                                                self.fitnesstst[np.argmax(self.fitnessval)], 
                                                self.complexity[np.argmax(self.fitnessval)], 
                                                self.population.population[np.argmax(self.fitnessval)]]


            if self.verbose == 2:
                print("\nStep 1. Fitness sorted")
                print(np.c_[self.fitnessval, self.fitnesstst, self.complexity, self.population.population][:10, :])
                # input("Press [enter] to continue")

            
            # Reorder models with ReRank function
            # -----------------------------------
            if self.rerank_error != 0.0 and self.iter >= self.iter_start_rerank:
                ord_rerank = self._rerank()
                self.population.population = self.population[ord_rerank]
                self.fitnessval = self.fitnessval[ord_rerank]
                self.fitnesstst = self.fitnesstst[ord_rerank]
                self.complexity = self.complexity[ord_rerank]
                _models = _models[ord_rerank]

                PopSorted = self.population._pop.copy()
                FitnessValSorted = self.fitnessval.copy()
                FitnessTstSorted = self.fitnesstst.copy()
                ComplexitySorted = self.complexity.copy()
                
                if self.verbose == 2:
                    print("\nStep 2. Fitness reranked")
                    print(np.c_[self.fitnessval, self.fitnesstst, self.complexity, self.population.population][:10, :])
                    # input("Press [enter] to continue")


            # Keep results
            # ---------------
            self._summary[iter, :] = parsimony_summary(self) # CAMBIAR AL CAMBIAR FUNCION PARSIMONY_SUMARY
            

            # Keep Best Solution
            # ------------------
            self.bestfitnessVal = self.fitnessval[0]
            self.bestfitnessTst = self.fitnesstst[0]
            self.bestcomplexity = self.complexity[0]
            self.bestsolution = np.concatenate([[self.bestfitnessVal, self.bestfitnessTst, self.bestcomplexity],self.population.population[0]])
            self.bestSolList.append(self.bestsolution)

            # Keep Best Model
            # ------------------
            self.best_model = _models[0]
            self.best_model_conf = self.population.getChromosome(0)
            

            # Keep elapsed time in minutes
            # ----------------------------
            tac = time.time()
            self.minutes_gen = (tac - tic) / 60.0
            self.minutes_total = self.minutes_total+self.minutes_gen
            

            # Keep this generation into the History list
            # ------------------------------------------
            if self.keep_history:
                self.history.append(pd.DataFrame(np.c_[self.population.population, self.fitnessval, self.fitnesstst, self.complexity], columns=list(self.population._params.keys())+self.population.colsnames+["fitnessval", "fitnesstst", "complexity"]))
            

            # Call to 'monitor' function
            # --------------------------
            if self.verbose > 0:
                parsimony_monitor(self)  
            
            if self.verbose == 2:
                print("\nStep 3. Fitness results")
                print(np.c_[self.fitnessval, self.fitnesstst, self.complexity, self.population.population][:10, :])
                # input("Press [enter] to continue")
            
            
            # Exit?
            # -----
            best_val_cost = self._summary[:,0][~np.isnan(self._summary[:,0])]
            if self.bestfitnessVal >= self.maxFitness:
                break
            if self.iter == self.maxiter:
                break
            if (len(best_val_cost)-(np.argmax(best_val_cost)+1)) >= self.early_stop:
                break
            
            
            # Selection Function
            # ------------------
            self._selection()

            if self.verbose == 2:
                print("\nStep 4. Selection")
                print(np.c_[self.fitnessval, self.fitnesstst, self.complexity, self.population.population][:10, :])
                # input("Press [enter] to continue")


            # CrossOver Function
            # ------------------
            if self.pcrossover > 0:
                nmating = int(np.floor(self.popSize/2))
                mating = np.random.choice(list(range(2 * nmating)), size=(2 * nmating), replace=False).reshape((nmating, 2))
                for i in range(nmating):
                    if self.pcrossover > np.random.uniform(low=0, high=1):
                        parents = mating[i, ]
                        self._crossover(parents=parents)
                if self.verbose == 2:
                    print("\nStep 5. CrossOver")
                    print(np.c_[self.fitnessval, self.fitnesstst, self.complexity, self.population.population][:10, :])
                    # input("Press [enter] to continue")


            # New generation with elitists
            # ----------------------------
            if (self.elitism > 0):
                self.population[:self.elitism] = PopSorted[:self.elitism]
                self.fitnessval[:self.elitism] = FitnessValSorted[:self.elitism]
                self.fitnesstst[:self.elitism] = FitnessTstSorted[:self.elitism]
                self.complexity[:self.elitism] = ComplexitySorted[:self.elitism]
            if (self.verbose == 2):
                print("\nStep 6. With Elitists")
                print(np.c_[self.fitnessval, self.fitnesstst, self.complexity, self.population.population][:10, :])
                # input("Press [enter] to continue")
            

            # Mutation function
            # -----------------
            if self.pmutation > 0:
                self._mutation()
                if self.verbose == 2:

                    print("\nStep 7. Mutation")
                    print(np.c_[self.fitnessval, self.fitnesstst, self.complexity, self.population.population][:10, :])
                    # input("Press [enter] to continue")
    
    def _rerank(self):
        r"""
        Function for reranking by complexity in parsimonious model selectionprocess. Promotes models with similar fitness but lower complexity to top positions.

        This method corresponds with the second step of parsimonious model selection (PMS) procedure.PMS works in the 
        following way: in each GA generation, best solutions are first sorted by their cost,J. Then, in a second step, 
        individuals with less complexity are moved to the top positions when theabsolute difference of their J is lower 
        than aobject@rerank_errorthreshold value. Therefore, theselection of less complex solutions among those with similar 
        accuracy promotes the evolution ofrobust solutions with better generalization capabilities.

        Returns
        -------
        numpy.array
            A vector with the new position of the individuals

        """

        cost1 = self.fitnessval.copy().astype(float)
        cost1[np.isnan(cost1)]= np.NINF

        sort = order(cost1, decreasing = True)
        cost1 = cost1[sort]
        complexity = self.complexity.copy()
        complexity[np.isnan(complexity)] = np.Inf
        complexity = complexity[sort]
        position = sort
  
        # start
        pos1 = 0
        pos2 = 1
        cambio = False
        error_posic = self.best_score
  
        while not pos1 == self.popSize:
            # Obtaining errors
            if pos2 >= self.popSize:
                if cambio:
                    pos2 = pos1+1
                    cambio = False
                else:
                    break
            error_indiv2 = cost1[pos2]
    
            # Compare error of first individual with error_posic. Is greater than threshold go to next point
            #      if ((Error.Indiv1-error_posic) > model@rerank_error) error_posic=Error.Indiv1
        
            error_dif = abs(error_indiv2-error_posic)
            if not np.isfinite(error_dif):
                error_dif = np.Inf
            if error_dif < self.rerank_error:
        
                # If there is not difference between errors swap if Size2nd < SizeFirst
                size_indiv1 = complexity[pos1]
                size_indiv2 = complexity[pos2]
                if size_indiv2<size_indiv1:
            
                    cambio = True
                
                    swap_indiv = cost1[pos1]
                    cost1[pos1] = cost1[pos2]
                    cost1[pos2] = swap_indiv
                            
                    complexity[pos1], complexity[pos2] = complexity[pos2], complexity[pos1]

                    position[pos1], position[pos2] = position[pos2], position[pos1]
                
                    if self.verbose == 2:
                        print(f"SWAP!!: pos1={pos1}({size_indiv1}), pos2={pos2}({size_indiv2}), error_dif={error_dif}")
                        print("-----------------------------------------------------")
                pos2 = pos2+1

            elif cambio:
                cambio = False
                pos2 = pos1+1
            else:
                pos1 = pos1+1
                pos2 = pos1+1
                error_dif2 = abs(cost1[pos1]-error_posic)
                if not np.isfinite(error_dif2):
                    error_dif2 = np.Inf
                if error_dif2 >= self.rerank_error:
                    error_posic = cost1[pos1]

        return position
    
    def _selection(self, *args, **kwargs):
        r"""
        Function for selection in GAparsimony.

        Functions implementing selection genetic operator in GA-PARSIMONY after parsimony_rerankprocess. 
        Linear-rank or Nonlinear-rank selection (Michalewicz (1996)). The type of selection is specified 
        with the model selection attribute, it can be: `linear`, `nlinear` or `random`. Modify the attributes: 
        `population`, `fitnessval`, `fitnesstst` and `complexity`.
        """

        # Establezco esta cabecera para permitir una reimplementación si se desea

        if self.selection == "linear":
            r = 2/(self.popSize*(self.popSize-1)) if "r" not in kwargs else kwargs["r"]
            q = 2/self.popSize if "q" not in kwargs else kwargs["q"]

            rank = range(self.popSize)
            prob = map(lambda x: q - (x)*r, rank)

            sel = np.random.choice(list(rank), size=self.popSize, replace=True, p=list(map(lambda x: np.min(np.ma.masked_array(np.array([max(0, x), 1]), np.isnan(np.array([max(0, x), 1])))), prob)))
        
        elif self.selection == "nlinear":
            # Nonlinear-rank selection
            # Michalewicz (1996) Genetic Algorithms + Data Structures = Evolution Programs. p. 60
            q = 0.25 if "q" not in kwargs else kwargs["q"]
            rank = list(range(self.popSize)) # population are sorted
            prob = np.array(list(map(lambda x: q*(1-q)**(x), rank)))
            prob = prob / prob.sum()
            
            sel = np.random.choice(list(rank), size=self.popSize, replace=True, p=list(map(lambda x: np.min(np.ma.masked_array(np.array([max(0, x), 1]), np.isnan(np.array([max(0, x), 1])))), prob)))
        
        elif self.selection == "random":
            sel = np.random.choice(list(range(self.popSize)), size=self.popSize, replace=True)
        
        else:
            raise Exception("Not a valid selection mode provided!!!!!!")
        
        self.population.population = self.population[sel]
        self.fitnessval = self.fitnessval[sel]
        self.fitnesstst = self.fitnesstst[sel]
        self.complexity = self.complexity[sel]
                
    def _crossover(self, parents, alpha=0.1, perc_to_swap=0.5):
        r"""
        Function for crossover in GAparsimony.

        Functions implementing particular crossover genetic operator for GA-PARSIMONY. 
        Method usesfor model parameters Heuristic Blending and random swapping for 
        binary selected features. Modify the attributes: `population`, `fitnessval`, 
        `fitnesstst` and `complexity`.

        Parameters
        ----------
        parents : list
            A list with two integers that correspond to the indices of the rows of the parents 
            of the current population.
        alpha : float, optional
            A tuning parameter for the Heuristic Blending outer bounds [Michalewicz, 1991].
            Typical and default value is `0.1`.
        perc_to_swap : float, optional
            Percentage of features for swapping in the crossovering process. Default value is `0.5`.
        """

        p=parents.copy()

        parents = self.population._pop[parents]
        children = parents # Vector
        pos_param_n = self.population._pos_n
        pos_param_c = self.population._pos_c
        pos_features = np.array(list(range(len(self.population._params), len(self.population._params) + len(self.population.colsnames))))
        
        # Heuristic Blending for parameters not categoricals
        Betas = np.random.uniform(size=len(pos_param_n), low=0, high=1)*(1+2*alpha)-alpha  # 1+alpha*2??????
        children[0,pos_param_n] = parents[0,pos_param_n]-Betas*parents[0,pos_param_n]+Betas*parents[1,pos_param_n]  ## MAP??
        children[1,pos_param_n] = parents[1,pos_param_n]-Betas*parents[1,pos_param_n]+Betas*parents[0,pos_param_n]

        # Random swapping for categorical parameters
        swap_param_c = np.random.uniform(size=len(pos_param_c), low=0, high=1)>=perc_to_swap
        if np.sum(swap_param_c)>0:
    
            parameters_c_parent1 = parents[0,pos_param_c]
            parameters_c_parent2 = parents[1,pos_param_c]
            pos_param_c = np.array(pos_param_c)[swap_param_c]
            children[0,pos_param_c] = parameters_c_parent2[swap_param_c]
            children[1,pos_param_c] = parameters_c_parent1[swap_param_c]
        
        # Random swapping for features
        swap_param = np.random.uniform(size=len(self.population.colsnames), low=0, high=1)>=perc_to_swap
        if np.sum(swap_param)>0:
    
            features_parent1 = parents[0,pos_features]
            features_parent2 = parents[1,pos_features]
            pos_features = pos_features[swap_param]
            children[0,pos_features] = features_parent2[swap_param]
            children[1,pos_features] = features_parent1[swap_param]
  
  
        # correct params that are outside (min and max)
        thereis_min = children[0] < self.population._min
        children[0,thereis_min] = self.population._min[thereis_min]
        thereis_min = children[1] < self.population._min
        children[1,thereis_min] = self.population._min[thereis_min]
        
        thereis_max = children[0] > self.population._max
        children[0,thereis_max] = self.population._max[thereis_max]
        thereis_max = (children[1] > self.population._max)
        children[1,thereis_max] = self.population._max[thereis_max]
  
        aux = np.empty(2)
        aux[:] = np.nan

        self.population[p] = children
        self.fitnessval[p] = aux.copy()
        self.fitnesstst[p] = aux.copy()
        self.complexity[p] = aux.copy()

    def _mutation(self):
        r"""
        Function for mutation in GAparsimony.

        Functions implementing mutation genetic operator for GA-PARSIMONY. Method mutes a `GAparsimony.pmutation` percentage 
        of them. If the value corresponds to a model parameter, algorithm uses uniform random mutation. For binary 
        select features, method sets to one if the random value between `[0,1]` is loweror equal to `GAparsimony.feat_mut_thres`. 
        Modify the attributes: `population`, `fitnessval`, `fitnesstst` and `complexity`.
        """


         # Uniform random mutation (except first individual)
        nparam_to_mute = round(self.pmutation*self.population.population.shape[1]*self.popSize)
        if nparam_to_mute<1:
            nparam_to_mute = 1
  
        for _ in range(nparam_to_mute):
  
            i = np.random.randint((self.not_muted), self.popSize)
            j = np.random.randint(0, self.population.population.shape[1])

            self.population[i,j] = self.population.random_gen[j](j, feat_mut_thres=self.feat_mut_thres)
            
            self.fitnessval[i] = np.nan
            self.fitnesstst[i] = np.nan
            self.complexity[i] = np.nan

    def _population(self, type_ini_pop="randomLHS"):
        r"""
        Population initialization in GA-PARSIMONY with a combined chromosome of model parameters 
        and selected features. Functions for creating an initial population to be used in the GA-PARSIMONY process.

        Generates a random population of `GAparsimony.popSize` individuals. For each individual a 
        random chromosome is generated with `len(GAparsimony.population._params)` real values in the `range[GAparsimony._min, GAparsimony._max] `
        plus `len(GAparsimony.population.colsnames)` random binary values for feature selection. `random` or Latin Hypercube Sampling can 
        be used to create a efficient spread initial population.

        Parameters
        ----------
        type_ini_pop : list, {'randomLHS', 'geneticLHS', 'improvedLHS', 'maximinLHS', 'optimumLHS'}, optional
            How to create the initial population. `random` optiom initialize a random population between the 
            predefined ranges. Values `randomLHS`, `geneticLHS`, `improvedLHS`, `maximinLHS` & `optimumLHS` 
            corresponds with several meth-ods of the Latin Hypercube Sampling (see `lhs` package for more details).

        Returns
        -------
        numpy.array
            A matrix of dimension `GAparsimony.popSize` rows and `len(GAparsimony.population._params)+len(GAparsimony.population.colsnames)` columns.

        """
  
        nvars = len(self.population._params) + len(self.population.colsnames)
        if type_ini_pop=="randomLHS":
            population = randomLHS(self.popSize, nvars, seed=self.seed_ini)
        elif type_ini_pop=="geneticLHS":
            population = geneticLHS(self.popSize, nvars, seed=self.seed_ini)
        elif type_ini_pop=="improvedLHS":
            population = improvedLHS(self.popSize, nvars, seed=self.seed_ini)
        elif type_ini_pop=="maximinLHS":
            population = maximinLHS(self.popSize, nvars, seed=self.seed_ini)
        elif type_ini_pop=="optimumLHS":
            population = optimumLHS(self.popSize, nvars, seed=self.seed_ini)
        elif type_ini_pop=="random":
            population = (np.random.rand(self.popSize*nvars) * (nvars - self.popSize) + self.popSize).reshape(self.popSize*nvars, 1)
  
        # Scale matrix with the parameters range
        population = population*(self.population._max-self.population._min)
        population = population+self.population._min
        # Convert features to binary 
        population[:, len(self.population._params):nvars] = population[:, len(self.population._params):nvars]<=self.feat_thres

        return population

    def __str__(self):
        dev =  "An object of class \"ga_parsimony\"\r" + \
                "Available slots:\r" + \
                f"bestfitnessVal: {self.bestfitnessVal}\r" + \
                f"bestfitnessTst: {self.bestfitnessTst}\r" + \
                f"bestcomplexity: {self.bestcomplexity}\r" + \
                f"bestsolution: {self.bestsolution}\r" + \
                f"min_param: {self.population._min}\r" + \
                f"max_param: {self.population._max}\r" + \
                f"nParams: {len(self.population._params)}\r" + \
                f"feat_thres: {self.feat_thres}\r" + \
                f"feat_mut_thres: {self.feat_mut_thres}\r" + \
                f"not_muted: {self.not_muted}\r" + \
                f"rerank_error: {self.rerank_error}\r" + \
                f"iter_start_rerank: {self.iter_start_rerank}\r" + \
                f"nFeatures: {len(self.population.colsnames)}\r" + \
                f"names_param: {list(self.population._params.keys())}\r" + \
                f"names_features: {self.population.colsnames}\r" + \
                f"popSize: {self.popSize}\r" + \
                f"iter: {self.iter}\r" + \
                f"early_stop: {self.early_stop}\r" + \
                f"maxiter: {self.maxiter}\r" + \
                f"minutes_gen: {self.minutes_gen}\r" + \
                f"minutes_total: {self.minutes_total}\r" + \
                f"suggestions: {self.suggestions}\r" + \
                f"population: {self.population}\r" + \
                f"elitism: {self.elitism}\r" + \
                f"pcrossover: {self.pcrossover}\r" + \
                f"pmutation: {self.pmutation}\r" + \
                f"best_score: {self.best_score}\r" + \
                f"solution_best_score: {self.solution_best_score}\r" + \
                f"fitnessval: {self.fitnessval}\r" + \
                f"fitnesstst: {self.fitnesstst}\r" + \
                f"complexity: {self.complexity}\r" + \
                f"summary: {self._summary}\r" + \
                f"bestSolList: {self.bestSolList}\r" + \
                f"history: \r"
        
        for h in self.history:
            dev += (h + "\r")

        return dev

    def __repr__(self):
        return self.__str__()

    def summary(self, **kwargs):
        r"""
        Summary for GA-PARSIMONY.

        Summary method for class `GAparsimony`. If it is assigned, it returns a dict if not displayed on the screen.

        Returns
        -------
        dict
            A `dict` with information about the GAparsimony object.

        """


        x = {"popSize" : self.popSize,
                "maxiter" : self.maxiter,
                "early_stop" : self.early_stop,
                "rerank_error" : self.rerank_error,
                "elitism" : self.elitism,
                "nParams" : len(self.population._params),
                "nFeatures" : len(self.population.colsnames),
                "pcrossover" : self.pcrossover,
                "pmutation" : self.pmutation,
                "feat_thres" : self.feat_thres,
                "feat_mut_thres" : self.feat_mut_thres,
                "not_muted" : self.not_muted,
                "domain" : np.stack([self.population._min, self.population._max], axis=0),
                "suggestions" : self.suggestions,
                "iter" : self.iter,
                "best_score" : self.best_score,
                "bestfitnessVal" : self.bestfitnessVal,
                "bestfitnessTst" : self.bestfitnessTst,
                "bestcomplexity" : self.bestcomplexity,
                "minutes_total" : self.minutes_total,
                "bestsolution" : self.bestsolution,
                "solution_best_score":self.solution_best_score}

        # Para contolar si lo está asignando
        try:
            frame = inspect.currentframe().f_back
            next_opcode = opcode.opname[frame.f_code.co_code[frame.f_lasti+2]]
            if next_opcode not in ["POP_TOP", "PRINT_EXPR"]: # Si no lo asigna
                return x 
        finally:
            del frame 


        print("+------------------------------------+")
        print("|             GA-PARSIMONY           |")
        print("+------------------------------------+\n")
        print("GA-PARSIMONY settings:")
        print(f" Number of Parameters      = {x['nParams']}")
        print(f" Number of Features        = {x['nFeatures']}")
        print(f" Population size           = {x['popSize']}")
        print(f" Maximum of generations    = {x['maxiter']}")
        print(f" Number of early-stop gen. = {x['early_stop']}")
        print(f" Elitism                   = {x['elitism']}")
        print(f" Crossover probability     = {x['pcrossover']}")
        print(f" Mutation probability      = {x['pmutation']}")
        print(f" Max diff(error) to ReRank = {x['rerank_error']}")
        print(f" Perc. of 1s in first popu.= {x['feat_thres']}")
        print(f" Prob. to be 1 in mutation = {x['feat_mut_thres']}")
        
        print("\n Search domain = ")
        print(pd.DataFrame(data=x["domain"], columns=list(self.population._params.keys())+self.population.colsnames, index=["Min_param", "Max_param"]))

        if x["suggestions"] is not None and x["suggestions"].shape[0]>0:
            print("Suggestions =")
            print(x["suggestions"])


        print("\n\nGA-PARSIMONY results:")
        print(f" Iterations                = {x['iter']+1}")
        print(f" Best validation score = {x['best_score']}")
        print(f"\n\nSolution with the best validation score in the whole GA process = \n")
        print(pd.DataFrame(data=x["solution_best_score"][np.newaxis,], columns=["fitnessVal","fitnessTst","complexity"]+list(self.population._params.keys())+self.population.colsnames))
        
        print(f"\n\nResults of the best individual at the last generation = \n")
        print(f" Best indiv's validat.cost = {x['bestfitnessVal']}")
        print(f" Best indiv's testing cost = {x['bestfitnessTst']}")
        print(f" Best indiv's complexity   = {x['bestcomplexity']}")
        print(f" Elapsed time in minutes   = {x['minutes_total']}")
        print(f"\n\nBEST SOLUTION = \n")
        print(pd.DataFrame(data=x["bestsolution"][np.newaxis,], columns=["fitnessVal","fitnessTst","complexity"]+list(self.population._params.keys())+self.population.colsnames))
    

    # Plot a boxplot evolution of val cost, tst cost and complexity for the elitists
    # ------------------------------------------------------------------------------
    def plot(self, min_iter=None, max_iter=None, main_label="Boxplot cost evolution", 
                                steps=5, size_plot=None, *args):
        r"""
        Plot of GA evolution of elitists.

        Plot method shows the evolution of validation and testing errors, and the number 
        of model features selected of elitists. White and grey box-plots represent validation 
        and testing errors of elitists evo-lution, respectively. Continuous and dashed-dotted 
        lines show the validation and testing error ofthe best individual for each generation, 
        respectively. Finally, the shaded area delimits the maximumand minimum number of 
        features, and the dashed line, the number of features of the best individual.

        Parameters
        ----------
        min_iter : int, optional
            Min GA iteration to visualize. Default `None`.
        max_iter : int, optional
            Max GA iteration to visualize.Default `None`.
        main_label : str, optional
            Main plot title.Default 'Boxplot cost evolution'.
        steps : int, optional
            Number of divisions in y-axis. Default `5`.
        size_plot : tuple, optional
            The size of the plot. Default `None`

        """

        if (len(self.history[0])<1):
            print("'GAparsimony.history' must be provided!! Set 'keep_history' to TRUE in ga_parsimony() function.")
        if not min_iter:
            min_iter = 0
        if not max_iter:
            max_iter = self.iter + 1
        
        # Preparacipon de los datos
        # ==========================
        mat_val = None
        mat_tst = None
        mat_complex = None
        for iter in range(min_iter, max_iter):
            mat_val = np.c_[mat_val, self.history[iter].fitnessval[:self.elitism]] if mat_val is not None else self.history[iter].fitnessval[:self.elitism]
            mat_tst = np.c_[mat_tst, self.history[iter].fitnesstst[:self.elitism]] if mat_tst is not None else self.history[iter].fitnesstst[:self.elitism]

            aux = np.sum(self.history[iter].values[:self.elitism,(len(self.population._params)):(len(self.population._params)+len(self.population.colsnames))], axis=1)
            mat_complex = np.c_[mat_complex, aux] if mat_complex is not None else aux

        x = list(range(min_iter, max_iter))
        mat_val = mat_val.astype(float)
        mat_tst = mat_tst.astype(float)
        mat_complex = mat_complex.astype(float)

        # Grafica
        # ======================
        fig, ax = plt.subplots() if not size_plot else plt.subplots(figsize=size_plot)

        sns.boxplot(x="x", y="y", data=pd.DataFrame(dict(x=np.repeat(range(min_iter, max_iter), self.elitism), y=mat_val.T.flatten())), color=(0.275191, 0.194905, 0.496005), width=0.4)
        sns.boxplot(x="x", y="y", data=pd.DataFrame(dict(x=np.repeat(range(min_iter, max_iter), self.elitism), y=mat_tst.T.flatten())), color=(0.626579, 0.854645, 0.223353), width=0.4)

        plt.suptitle(main_label, fontsize=16)
        plt.title(f"Results for the last best individual: Val={round(self.bestfitnessVal, 5)}, Test={round(self.bestfitnessTst, 5)}, Num.Features={int(mat_complex[0, -1])}")

        # Eje de la derecha
        sns.lineplot(x=x, y=mat_val[0], color=(0.153364, 0.497, 0.557724), style=True, dashes=[(10, 2)])
        ax.legend([],[], frameon=False)

        sns.lineplot(x=x, y=mat_tst[0], color=(0.122312, 0.633153, 0.530398), style=True, dashes=[(2, 2, 10, 2)]) # 2pt line, 2pt break, 10pt line, 2pt break
        ax.legend([],[], frameon=False)
        
        
        ax.set_ylabel("Metric")
        
        # Eje de la izquierda
        ax2 = plt.twinx()
        sns.lineplot(x=x, y=mat_complex[0], color=(0.212395, 0.359683, 0.55171))
        

        plt.fill_between(x = x,
                 y1 = np.min(mat_complex.T, axis=1),
                 y2 = np.max(mat_complex.T, axis=1),
                 alpha = 0.1,
                 facecolor = (0.212395, 0.359683, 0.55171))

        plt.ylabel("Number of Features of Best Indiv.")

        ax.legend(handles=[mlines.Line2D([], [], linestyle="--", color=(0.153364, 0.497, 0.557724), label='Validation METRIC of best individual'),
                            mlines.Line2D([], [], linestyle="-.", color=(0.122312, 0.633153, 0.530398), label='Testing METRIC of best individual'),
                            mlines.Line2D([], [], linestyle="-", color=(0.212395, 0.359683, 0.55171), label='Number of features of best individual'),
                            mpatches.Patch(color=(0.275191, 0.194905, 0.496005), label='Validation metric'),
                            mpatches.Patch(color=(0.626579, 0.854645, 0.223353), label='Testing metric')],
                            loc=3, # 3 o 1
                            ncol=2, borderaxespad=0.)
        
        plt.xticks(np.arange(0,max_iter, steps))
        ax.set_xlabel("Number of Generation")

        ax.set_zorder(ax2.get_zorder()+1) 
        ax.patch.set_visible(False)

        plt.show()        


    def importance(self):
        r"""
        Percentage of appearance of each feature in elitist population.

        Shows the percentage of appearance of each feature in the whole GA-PARSIMONY 
        process butonly for the elitist-population. If it is assigned, it returns a 
        dict if not displayed on the screen.

        Returns
        -------
        numpy.array
            A `numpy.array` with information about feature importance.

        """
    
        if len(self.history[0]) < 1:
            print("'object.history' must be provided!! Set 'keep_history' to TRUE in ga_parsimony() function.")
        min_iter = 1
        max_iter = self.iter
        
        nelitistm = self.elitism
        features_hist = None
        for iter in range(min_iter, max_iter+1):
            data = self.history[iter].iloc[:nelitistm, len(self.population._params):-3].values
            features_hist = data if features_hist is None else np.r_[features_hist, data] ## ANALIZAR CON CUIDADO

        importance = np.mean(features_hist, axis=0)
        sort = order(importance,decreasing = True)
        imp_features = pd.DataFrame(100*importance[order(importance,decreasing = True)][np.newaxis], columns=self.history[0].columns[len(self.population._params):-3][sort])

        # Para contolar si lo está asignando
        try:
            frame = inspect.currentframe().f_back
            next_opcode = opcode.opname[frame.f_code.co_code[frame.f_lasti+2]]
            if next_opcode not in ["POP_TOP", "PRINT_EXPR"]: # Si no lo asigna
                return imp_features 
        finally:
            del frame 

        print("+--------------------------------------------+")
        print("|                  GA-PARSIMONY              |")
        print("+--------------------------------------------+\n")
        print("Percentage of appearance of each feature in elitists: \n")
        print(imp_features)

    def predict(self, X):
        r"""
        Predict result for samples in X.

        Parameters
        ----------
        X : numpy.array or pandas.DataFrame
            Samples.

        Returns
        -------
        numpy.array
            A `numpy.array` with predictions.

        """

        X = X.values if type(X).__name__ == 'DataFrame' else X
        return self.best_model.predict(X[:, self.best_model_conf.columns.astype(bool)])








# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
import os

# For installing PyTorch and Torchvision in Windows
import sys
import subprocess

# Get the long description from the README file
with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

def parse_requirements(filename):
    """ load requirements from a pip requirements file """
    lineiter = (line.strip() for line in open(filename, 'r'))
    return [line for line in lineiter if line and not line.startswith("#")]


install_reqs = parse_requirements("requirements.txt")

def remove_requirements(requirements, remove_elem):
    new_requirements = []
    for requirement in requirements:
        if remove_elem not in requirement:
            new_requirements.append(requirement)

    return new_requirements

version = None
with open('GAparsimony/__init__.py', 'r', encoding='utf-8') as f:

    version = f.readline().split('=')[-1].strip().replace('"','')



# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    # This is the name of your project. The first time you publish this
    # package, this name will be registered for you. It will determine how
    # users can install this project, e.g.:
    #
    # $ pip install sampleproject
    #
    # And where it will live on PyPI: https://pypi.org/project/sampleproject/
    #
    # There are some restrictions on what makes a valid project name
    # specification here:
    # https://packaging.python.org/specifications/core-metadata/#name
    name='GAparsimony',  # Required
    python_requires='>=3.7',

    # Versions should comply with PEP 440:
    # https://www.python.org/dev/peps/pep-0440/
    #
    # For a discussion on single-sourcing the version across setup.py and the
    # project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=version,  # Required

    # This is a one-line description or tagline of what your project does. This
    # corresponds to the "Summary" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#summary
    description="Combines feature selection, model tuning, and parsimonious model selection with GA optimization. GA selection procedure is based on separate cost and complexity evaluations. Therefore, the best individuals are initially sorted by an error fitness function, and afterwards, models with similar costs are rearranged according to modelcomplexity measurement so as to foster models of lesser complexity. The algorithm can be run sequentially or in parallel.",  # Required

    # This is an optional longer description of your project that represents
    # the body of text which users will see when they visit PyPI.
    #
    # Often, this is the same as your README, so you can just read it in from
    # that file directly (as we have already done above)
    #
    # This field corresponds to the "Description" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#description-optional
    long_description=long_description,  # Optional
    long_description_content_type="text/markdown",

    # This should be a valid link to your project's main homepage.
    #
    # This field corresponds to the "Home-Page" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#home-page-optional
    url='https://github.com/misantam/GAparsimony',  # Optional

    # This should be your name or the name of the organization which owns the
    # project.
    author='Millán Santamaría ',  # Optional

    # This should be a valid email address corresponding to the author listed
    # above.
    author_email='GAparsimony@gmail.com',  # Optional

    maintainer='Millán Santamaría',
    maintainer_email='GAparsimony@gmail.com',

    license="MIT License",
    # Classifiers help users find your project by categorizing it.
    #
    # For a list of valid classifiers, see
    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',       

        # Topics
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',


        # Pick your license as you wish
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3'
    ],

    # This field adds keywords for your project which will appear on the
    # project page. What does your project relate to?
    #
    # Note that this is a string of words separated by whitespace, not a list.
    # keywords='instance semantic segmentation pytorch fastai fastai2 fastaiv2 saliend object detection',  # Optional

    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().
    #
    # Alternatively, if you just want to distribute a single Python file, use
    # the `py_modules` argument instead as follows, which will expect a file
    # called `my_module.py` to exist:
    #
    #   py_modules=["my_module"],
    #
    packages=find_packages(),  # Required

    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip's requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=install_reqs,  # Optional

    # List additional groups of dependencies here (e.g. development
    # dependencies). Users will be able to install these using the "extras"
    # syntax, for example:
    #
    #   $ pip install sampleproject[dev]
    #
    # Similar to `install_requires` above, these must be valid existing
    # projects.
    #extras_require={  # Optional
    #    'dev': ['check-manifest'],
    #    'test': ['coverage'],
    #},

    # If there are data files included in your packages that need to be
    # installed, specify them here.
    #
    # If using Python 2.6 or earlier, then these have to be included in
    # MANIFEST.in as well.
    include_package_data=True,
    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files
    #
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    #data_files=[('my_data', ['data/data_file'])],  # Optional

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # `pip` to create the appropriate form of executable for the target
    # platform.
    # For example, the following would provide a command called `sample` which
    # executes the function `main` from this package when invoked:
)
